#ifndef ATREE_SET_H
#define ATREE_SET_H

#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "misc/thread_includes.h"
#include "misc/padded_types.h"
#include "data_structures/oo_set_interface.h"
#include "locks/drmcs_lock.h"


void custom_free2(void * data);
void * custom_malloc2(int size);

#ifndef ATREE_MALLOC
#      define ATREE_MALLOC(size) malloc(size)
#endif

#ifndef ATREE_FREE
#      define ATREE_FREE(data) free(data)
#endif

#define ATREE_MAX_CONTENTION_STATISTICS 20
#define ATREE_MIN_CONTENTION_STATISTICS -20
#define ATREE_LOCK_SUCCESS_STATS_CONTRIB -1
#define ATREE_LOCK_FAILURE_STATS_CONTRIB 2
#define ATREE_REBUILD_CONSTANT 2

#define ATREE_LOCK_TYPE pthread_mutex_t
#define ATREE_LOCK_INITIALIZER(lock_ptr)                                \
    pthread_mutexattr_t attr;                                           \
    pthread_mutexattr_init(&attr);                                      \
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);         \
    pthread_mutex_init(lock_ptr, &attr);                                \
    pthread_mutexattr_destroy(&attr);
#define ATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) (!pthread_mutex_trylock(lock_ptr))
#define ATREE_LOCK_LOCK_FUNCTION(lock_ptr) pthread_mutex_lock(lock_ptr)
#define ATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) pthread_mutex_unlock(lock_ptr)

typedef struct {
    ATREE_LOCK_TYPE lock;
    int validation;
    int statistics;
} ATreeLock;

typedef struct ValueNodeImpl{
    struct ValueNodeImpl * left;
    struct ValueNodeImpl * right;
    unsigned int valueSize;
    bool deleted;
    void * value;
} ValueNode;

typedef struct {
    struct ValueNodeImpl * root;
    ATreeLock lock;
    bool valid;
    unsigned int nrOfElements;
    unsigned int nrOfDeletedElements;
} BaseNode;

typedef struct {
    void * left;
    void * right;
    unsigned int valueSize;
    void * value;
} RouteNode;

typedef union {
    RouteNode route;
    BaseNode base;
} BaseOrRouteNodeContainer;

typedef struct {
    bool isBaseNode;
    BaseOrRouteNodeContainer baseOrRoute;
} BaseOrRouteNode;

typedef struct {
    DRMCSLock globalLock;
    BaseOrRouteNode * root;
    unsigned int keyPosition;
    void * (*extract_key)(void * v, int keyPos);
    unsigned int (*hash_key)(void * k);
    bool (*are_equal)(void * v1, void * v2);
    bool (*less_than)(void * v1, void * v2);
    char * (*to_string)(void * v1);
    void (*free)(void * v, void * context);
    void * freeContext;
} ATreeSet;

extern
_Alignas(CACHE_LINE_SIZE)
OOSetMethodTable ATREE_SET_METHOD_TABLE;

static inline
void atree_lock_init(ATreeLock * lock){
    lock->statistics = 0;
    lock->validation = 0;
    ATREE_LOCK_INITIALIZER(&lock->lock);
}

static inline
void atree_lock(ATreeLock * lock){
    int failureAccumelator = 0;
    for(int i = 0; i < 1000; i++){
        if(ATREE_LOCK_TRY_LOCK_FUNCTION(&lock->lock)){
            lock->statistics += ATREE_LOCK_SUCCESS_STATS_CONTRIB;
            lock->statistics += failureAccumelator;
            if(lock->statistics < ATREE_MIN_CONTENTION_STATISTICS){
                lock->statistics = 0;
            }
            lock->validation++;
            assert(lock->validation == 1);
            return;
        }else{
            failureAccumelator += ATREE_LOCK_FAILURE_STATS_CONTRIB;
            thread_yield();
        }
    }
    ATREE_LOCK_LOCK_FUNCTION(&lock->lock);
    lock->statistics += ATREE_LOCK_SUCCESS_STATS_CONTRIB;
    lock->statistics += failureAccumelator;
    if(lock->statistics < ATREE_MIN_CONTENTION_STATISTICS){
        lock->statistics = 0;
    }
    lock->validation++;
    assert(lock->validation == 1);
}

static inline
void atree_unlock(ATreeLock * lock){
    lock->validation--;
    assert(lock->validation == 0);
    ATREE_LOCK_UNLOCK_FUNCTION(&lock->lock);
}

static inline
BaseOrRouteNode * allocateBaseNode(){
    BaseOrRouteNode * newBase = ATREE_MALLOC(sizeof(BaseOrRouteNode));
    newBase->isBaseNode = true;
    atree_lock_init(&newBase->baseOrRoute.base.lock);
    newBase->baseOrRoute.base.root = NULL;
    newBase->baseOrRoute.base.nrOfElements = 0;
    newBase->baseOrRoute.base.nrOfDeletedElements = 0;
    newBase->baseOrRoute.base.valid = true;
    return newBase;
}

static inline
BaseOrRouteNode * allocateRouteNode(unsigned int valueSize,
                                    void * value,
                                    BaseOrRouteNode * newBaseLeft,
                                    BaseOrRouteNode * newBaseRight){
    BaseOrRouteNode * newRoute = ATREE_MALLOC(sizeof(BaseOrRouteNode));
    newRoute->isBaseNode = false;
    newRoute->baseOrRoute.route.left = newBaseLeft;
    newRoute->baseOrRoute.route.right = newBaseRight;
    newRoute->baseOrRoute.route.valueSize = valueSize;
    newRoute->baseOrRoute.route.value = ATREE_MALLOC(valueSize);
    memcpy(newRoute->baseOrRoute.route.value, value, valueSize);
    return newRoute;
}



static inline
ValueNode * allocateValueNode(unsigned int size, void * value, bool copyIn){
    ValueNode * newValueNode = ATREE_MALLOC(sizeof(ValueNode));
    newValueNode->left = NULL;
    newValueNode->right = NULL;
    newValueNode->valueSize = size;
    newValueNode->deleted = false;
    if(copyIn){
        newValueNode->value = ATREE_MALLOC(size);
        memcpy(newValueNode->value, value, size);
    } else {
        newValueNode->value = value;
    }
    return newValueNode;
}

  /* private var root:Node = BaseNode(null) */

  /* def printRootStructure() = printStructure(root) */

  /* def printStructure(currentNode:Node):String = currentNode match { */
  /*   case BaseNode(root, _, _, nrOfElements) => "B("+ (if(root == null) null else(root.key)) + ","+nrOfElements+")" */
  /*   case RouteNode(key, left, right) => "R("+key+","+printStructure(left)+","+printStructure(right)+")" */
  /* } */


//http://www.hackersdelight.org/hdcodetxt/nlz.c.txt
static inline
unsigned int nlz1(unsigned int x) {
   int n;

   if (x == 0) return(32);
   n = 0;
   if (x <= 0x0000FFFF) {n = n +16; x = x <<16;}
   if (x <= 0x00FFFFFF) {n = n + 8; x = x << 8;}
   if (x <= 0x0FFFFFFF) {n = n + 4; x = x << 4;}
   if (x <= 0x3FFFFFFF) {n = n + 2; x = x << 2;}
   if (x <= 0x7FFFFFFF) {n = n + 1;}
   return n;
}

static inline
unsigned int log2nlz(unsigned int bits) {
    if( bits == 0 ) {
        return 0; 
    }else {
        return 31 - nlz1( bits );
    }
}

 static inline void at_standard_free(void * toFree, void * context){
     (void)context;
     ATREE_FREE(toFree);
 }

static inline
void plain_at_set_init(ATreeSet * set,
                       unsigned int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       unsigned int (*hash_key)(void * k),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       char * (*to_string)(void * v1),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    drmcs_initialize(&set->globalLock);
    set->keyPosition = keyPosition;
    set->extract_key = extract_key;
    set->hash_key = hash_key;
    set->are_equal = are_equal;
    set->less_than = less_than;
    set->to_string = to_string;
    if(custom_free == NULL){
        set->free = at_standard_free;
        set->freeContext = NULL;
    }else{
        set->free = custom_free;
        set->freeContext = freeContext;
    }
    set->root = allocateBaseNode();
}


static inline
ATreeSet * plain_at_set_create(unsigned int keyPosition,
                               void * (*extract_key)(void * v, int keyPos),
                               unsigned int (*hash_key)(void * k),
                               bool (*are_equal)(void * v1, void * v2),
                               bool (*less_than)(void * v1, void * v2),
                               char * (*to_string)(void * v1)){
    ATreeSet * set = ATREE_MALLOC(sizeof(ATreeSet));
    plain_at_set_init(set,
                      keyPosition,
                      extract_key,
                      hash_key,
                      are_equal,
                      less_than,
                      to_string,
                      NULL,
                      NULL);
    return set;
}

static inline
int compare_hash_codes2(unsigned int code1, unsigned int code2){
    return code1 - code2;
}

static inline
unsigned int countNodes(ValueNode * node){ 
    if(node == NULL) return 0;
    else if(node->left == NULL ) return countNodes(node->right) + 1;
    else if(node->right == NULL ) return countNodes(node->left) + 1;
    else return countNodes(node->left) + 1 + countNodes(node->right);
}

static inline
ValueNode * flattern(ValueNode * currentNode, ValueNode ** startPointWriteBack) {
    if(*startPointWriteBack == NULL && currentNode->left == NULL){
        *startPointWriteBack = currentNode;
    }
    if(currentNode->left == NULL && currentNode->right == NULL){
        return currentNode;
    }else if (currentNode->right == NULL){
        ValueNode * rightEnd = flattern(currentNode->left, startPointWriteBack);
        rightEnd->right = currentNode;
        return currentNode;
    } else if(currentNode->left == NULL){
        ValueNode * oldStartPoint = *startPointWriteBack;
        *startPointWriteBack = NULL;
        ValueNode * rightEnd = flattern(currentNode->right, startPointWriteBack);
        currentNode->right = *startPointWriteBack;
        *startPointWriteBack = oldStartPoint;
        return rightEnd;
    } else {
        ValueNode * rightEnd = flattern(currentNode->left, startPointWriteBack);
        rightEnd->right = currentNode;
        ValueNode *  oldStartPoint = *startPointWriteBack;
        *startPointWriteBack = NULL;
        ValueNode * rightEnd2 = flattern(currentNode->right, startPointWriteBack);
        currentNode->right = *startPointWriteBack;
        *startPointWriteBack = oldStartPoint;
        return rightEnd2;
    }
  }

static inline
unsigned int removeDeletedFromFlat(ValueNode ** startInOut,
                                   void (*custom_free)(void * v, void * context),
                                   void * freeContext) {
    ValueNode * pseudoStartNode = ATREE_MALLOC(sizeof(ValueNode));//TODO Make more efficent
    pseudoStartNode->right = *startInOut;
    ValueNode * prevNode = pseudoStartNode;
    ValueNode * currentNode = *startInOut;
    unsigned int nrOfRemovedNodes = 0;
    while(currentNode != NULL){
        if(currentNode->deleted){
            custom_free(currentNode->value, freeContext);
            prevNode->right = currentNode->right;
            nrOfRemovedNodes = nrOfRemovedNodes + 1;
            ValueNode * oldCurrentNode = currentNode;
            currentNode = currentNode->right;
            ATREE_FREE(oldCurrentNode);
        }else{
            prevNode = currentNode;
            currentNode = currentNode->right;
        }
    }
    *startInOut = pseudoStartNode->right;
    ATREE_FREE(pseudoStartNode);
    return nrOfRemovedNodes;
}

static inline
ValueNode * rebuildAtFlat(ValueNode **currentNode, unsigned int fromIndex, unsigned int toIndex) {
    unsigned int difference = toIndex - fromIndex;
    if(difference == 0) {
        return NULL;
    } else if(difference == 1) {
        ValueNode * savedCurrentNode = *currentNode;
        *currentNode = (*currentNode)->right;
        savedCurrentNode->left = NULL;
        savedCurrentNode->right = NULL;
        return savedCurrentNode;
    } else {
        unsigned int firstHalf = difference / 2;
        ValueNode * firstHalfNodes = rebuildAtFlat(currentNode, fromIndex, fromIndex + firstHalf);
        ValueNode * center = *currentNode;
        *currentNode = (*currentNode)->right;
        ValueNode * secondHalfNodes = rebuildAtFlat(currentNode, fromIndex + firstHalf + 1, toIndex);
        center->left = firstHalfNodes;
        center->right = secondHalfNodes;
        return center;
    }
}

/* static inline */
/* void at_set_print_helper_helper2(ValueNode * node, char * (*to_string)(void * v1)){ */
/*     if(node==NULL){ */
/*         printf("NULL"); */
/*     }else{ */
/*         printf("V("); */
/*         char * valueString = to_string(node->value); */
/*         printf("%s", valueString); */
/*         free(valueString); */
/*         printf(","); */
/*         at_set_print_helper_helper2(node->left, to_string); */
/*         printf(","); */
/*         at_set_print_helper_helper2(node->right, to_string); */
/*         printf(")"); */
/*     } */
/* } */
/* static */
/* char * debug_my_to_string(void * value){ */
/*     int  valuei = *((int *)value); */
/*     char * str = malloc(15); */
/*     sprintf(str, "%d", valuei); */
/*     return str; */
/* } */

static inline
ValueNode * rebuildAt(ValueNode * currentNode,
                      unsigned int countFromBranch,
                      BaseNode * baseNode,
                      void (*custom_free)(void * v, void * context),
                      void * freeContext) {
    ValueNode * startPointWriteBackDest = NULL;
    ValueNode ** startPointWriteBack = &startPointWriteBackDest;
    flattern(currentNode, startPointWriteBack);
    unsigned int nrOfRemovedNodes = removeDeletedFromFlat(startPointWriteBack, custom_free, freeContext);
    baseNode->nrOfElements = baseNode->nrOfElements - nrOfRemovedNodes;
    baseNode->nrOfDeletedElements = baseNode->nrOfElements - baseNode->nrOfDeletedElements;
    if((*startPointWriteBack) != NULL){
        return rebuildAtFlat(startPointWriteBack, 0, countFromBranch - nrOfRemovedNodes);
    } else { 
        return NULL;
    }
}


static inline
void rebuildOnPath(void * key, 
                   BaseNode * baseNode,
                   int pathLengthO,
                   int keyPosition,
                   void * (*extract_key)(void * v, int keyPos),
                   bool (*are_equal)(void * v1, void * v2),
                   bool (*less_than)(void * v1, void * v2),
                   void (*custom_free)(void * v, void * context),
                   void * freeContext) {
    ValueNode * currentNode = baseNode->root;
    ValueNode* pathStack[pathLengthO];
    int level = 0;
    while(!are_equal(extract_key(currentNode->value, keyPosition), key)) {
        pathStack[level] = currentNode;
        level = level + 1;
        //TODO no need to extract the same key twice
        if(less_than(key, extract_key(currentNode->value, keyPosition))){
            currentNode = currentNode->left;
        } else {
            currentNode = currentNode->right;
        }
    }
    unsigned int countFromBranch = 1;
    level = level - 1;
    ValueNode * prevNode = currentNode;
    unsigned int pathLength = 1;
    ValueNode * newCurrentNode = NULL;
    ValueNode ** base = NULL;
    bool rebuildDone = false;
    while(level != -1) {
        prevNode = currentNode;
        currentNode = pathStack[level];
        ValueNode * notFromBranch = NULL;
        if(currentNode->left == prevNode){
            notFromBranch = currentNode->right;
        } else {
            notFromBranch = currentNode->left;
        }
        countFromBranch = countFromBranch + 1 + countNodes(notFromBranch);
        if(pathLength > (ATREE_REBUILD_CONSTANT * log2nlz( countFromBranch ))){
            newCurrentNode = rebuildAt(currentNode, countFromBranch, baseNode, custom_free, freeContext);
            if(level == 0) {
                baseNode->nrOfElements = countFromBranch;
                base = &baseNode->root;
            } else {
                ValueNode * end = pathStack[level - 1];
                if(end->left == currentNode){
                    base = &end->left;
                }else{
                    base = &end->right;
                }
            }
            level = -1;
            rebuildDone = true;
        } else {
            level = level - 1;
            pathLength = pathLength + 1;
        }
    }
    if(!rebuildDone){
        baseNode->nrOfElements = countFromBranch;
        newCurrentNode = rebuildAt(baseNode->root, countFromBranch, baseNode, custom_free, freeContext);
        base = &baseNode->root;
    }
    
    *base = newCurrentNode;
}

static inline
void insertValueNode(ValueNode * node, 
                     ValueNode * nodeToInsert,
                     int keyPosition,
                     void * (*extract_key)(void * v, int keyPos),
                     bool (*less_than)(void * v1, void * v2)){
    ValueNode * currentNode = node;
    void * nodeToInsertKey = extract_key(nodeToInsert->value, keyPosition);
    while(true){
        if(less_than(nodeToInsertKey,  extract_key(currentNode->value, keyPosition))) {
            if(currentNode->left == NULL){
                //Insert here
                currentNode->left = nodeToInsert;
                return;
            }else {
                currentNode = currentNode->left;
            }
        } else {
            if(currentNode->right == NULL){
                //Insert here
                currentNode->right = nodeToInsert;
                return;
            }else {
                currentNode = currentNode->right;
            }
        }
    }
}


static inline
bool contentionReduceSplit(BaseOrRouteNode *  baseNodeContainer,
                           void ** parent,
                           ATreeSet * set,
                           int keyPosition,
                           void * (*extract_key)(void * v, int keyPos),
                           bool (*less_than)(void * v1, void * v2)) {
    BaseNode * node = &baseNodeContainer->baseOrRoute.base;
    if(node == NULL || node->root == NULL || (node->root->left == NULL && node->root->right == NULL)){
        node->lock.statistics = 0;
        return false;
    }
    ValueNode * splitValueNode = NULL;
      if (node->root->left == NULL) {
        //To get non empty split
          ValueNode * insertTo = node->root->right;
          node->root->right = NULL;
          insertValueNode(insertTo,
                          node->root,
                          keyPosition,
                          extract_key,
                          less_than);
          splitValueNode = insertTo;
      }else{
          splitValueNode = node->root;
      }

      void * splitKey = splitValueNode->value;
      unsigned int splitKeyValueSize = splitValueNode->valueSize;
      ValueNode * leftBaseRoot = splitValueNode->left;
      splitValueNode->left = NULL;
      ValueNode * rightBaseRoot = NULL;
      if(splitValueNode->right == NULL)
          rightBaseRoot = splitValueNode;
      else {
          //To narrow down faster
          ValueNode * base = splitValueNode->right;
          splitValueNode->right = NULL;
          insertValueNode(base,
                          splitValueNode,
                          keyPosition,
                          extract_key,
                          less_than);
          rightBaseRoot = base;
      }
      //We can not use the same because it might contain less than what a thread waiting for expect
      BaseOrRouteNode * newBaseRight = allocateBaseNode();
      BaseNode * rightBCont = &newBaseRight->baseOrRoute.base;
      rightBCont->root = rightBaseRoot;
      rightBCont->nrOfElements = node->nrOfElements / 2;
      rightBCont->nrOfDeletedElements = node->nrOfDeletedElements / 2;

      BaseOrRouteNode * newBaseLeft = allocateBaseNode();
      BaseNode * leftBCont = &newBaseLeft->baseOrRoute.base;
      leftBCont->root = leftBaseRoot;
      leftBCont->nrOfElements = node->nrOfElements / 2;
      leftBCont->nrOfDeletedElements = node->nrOfDeletedElements / 2;
      
      BaseOrRouteNode * newRoute = allocateRouteNode(splitKeyValueSize,
                                                     splitKey,
                                                     newBaseLeft,
                                                     newBaseRight);
      node->valid = false;
      atree_unlock(&node->lock);
      drmcs_runlock(&set->globalLock);
      drmcs_lock(&set->globalLock); // potentially this could be done much more rearly
      *parent = newRoute;
      drmcs_unlock(&set->globalLock);
      pthread_mutex_destroy(&node->lock.lock); //TODO do we need this?
      ATREE_FREE(baseNodeContainer);
      drmcs_rlock(&set->globalLock);
      return true;
}



static inline
bool at_update_helper_ValueNode(ValueNode * searchNode,
                                BaseNode * baseNode,
                                void * valuePtr,
                                unsigned int valueSize,
                                int keyPosition,
                                bool overwrite,
                                void * (*extract_key)(void * v, int keyPos),
                                bool (*are_equal)(void * v1, void * v2),
                                bool (*less_than)(void * v1, void * v2),
                                bool copyIn,
                                void ** replacedWriteBack,
                                void (*custom_free)(void * v, void * context),
                                void * freeContext){
    void * key = extract_key(valuePtr, keyPosition);
    ValueNode * currentNode = searchNode;
    ValueNode * prevNode = NULL;
    bool lastLeft = true;
    int pathLength = -1;
    while(currentNode != NULL){
        void * nodeKey = extract_key(currentNode->value, keyPosition);
        prevNode = currentNode;
        pathLength = pathLength + 1;
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->left;
            lastLeft = true;
        } else if (are_equal(key, nodeKey)) {
            //Replace
            if(currentNode->deleted){//Delted
                baseNode->nrOfDeletedElements = baseNode->nrOfDeletedElements - 1;
                if(replacedWriteBack == NULL){
                    custom_free(currentNode->value, freeContext);
                }else{
                    *replacedWriteBack = currentNode->value;
                }
                if(copyIn){
                    currentNode->value = ATREE_MALLOC(valueSize);
                    memcpy(currentNode->value, valuePtr, valueSize);
                }else{
                    currentNode->value = valuePtr;
                }
                currentNode->deleted = false;
                return true;
            }else if(overwrite){
                if(replacedWriteBack == NULL){
                    custom_free(currentNode->value, freeContext);
                }else{
                    *replacedWriteBack = currentNode->value;
                }
                if(copyIn){
                    currentNode->value = ATREE_MALLOC(valueSize);
                    memcpy(currentNode->value, valuePtr, valueSize);
                }else{
                    currentNode->value = valuePtr;
                }
                currentNode->valueSize = valueSize;
                return false;
            }else{
                return false;
            }
        } else {
            currentNode = currentNode->right;
            lastLeft = false;
        }
    }
    //Insert new node
    baseNode->nrOfElements = baseNode->nrOfElements + 1;
    ValueNode * newNode = allocateValueNode(valueSize, valuePtr, copyIn);
    if(lastLeft) {
        prevNode->left = newNode;
    } else {
        prevNode->right = newNode;
    }
    //TODO find better solution to this
    if(pathLength > (int)(ATREE_REBUILD_CONSTANT * log2nlz( baseNode->nrOfElements + baseNode->nrOfDeletedElements ))){
        rebuildOnPath(key, baseNode, pathLength + 1, keyPosition,
                      extract_key,
                      are_equal,
                      less_than,
                      custom_free,
                      freeContext);
    }    
    return true;
}

static inline
bool at_set_insert_opt(ATreeSet * set,
                       void * valuePtr,
                       unsigned int valueSize,
                       int keyPosition,
                       bool overwrite,
                       void * (*extract_key)(void * v, int keyPos),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       bool copyIn,
                       void ** replacedWriteBack,
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    void * key = extract_key(valuePtr, keyPosition);
    //Find base node
    BaseOrRouteNode * currentNode = set->root;
    RouteNode * currentRouteNode = NULL;
    while(!currentNode->isBaseNode){
        currentRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentRouteNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentRouteNode->left;
        } else {
            currentNode = currentRouteNode->right;
        }
    }
    //Lock and handle contention
    BaseNode * node = &currentNode->baseOrRoute.base;
    ATreeLock * lock = &node->lock;
    void ** parent = NULL;
    atree_lock(lock);
    if(!node->valid){
        //Retry
        atree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        return at_set_insert_opt(set,
                                 valuePtr,
                                 valueSize,
                                 keyPosition,
                                 overwrite,
                                 extract_key,
                                 are_equal,
                                 less_than,
                                 copyIn,
                                 replacedWriteBack,
                                 custom_free,
                                 freeContext);
    } else if(lock->statistics > ATREE_MAX_CONTENTION_STATISTICS){
        if(currentRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentRouteNode->left == currentNode){
            parent = (void**)&currentRouteNode->left;
        }else{
            parent = (void**)&currentRouteNode->right;
        }
    }
    //Handle empty before and then normal case
    bool returnValue = false;
    if(node->root == NULL) {
        ValueNode * newNode = allocateValueNode(valueSize, valuePtr, copyIn);
        node->root = newNode;
        node->nrOfElements = node->nrOfElements + 1;
        returnValue = true;
    } else {
        returnValue = at_update_helper_ValueNode(node->root,
                                                 node,
                                                 valuePtr,
                                                 valueSize,
                                                 keyPosition,
                                                 overwrite,
                                                 extract_key,
                                                 are_equal,
                                                 less_than,
                                                 copyIn,
                                                 replacedWriteBack,
                                                 custom_free,
                                                 freeContext);
    }
    if(parent != NULL){
        if(!contentionReduceSplit(currentNode,
                                  parent,
                                  set,
                                  keyPosition,
                                  extract_key,
                                  less_than)){
            atree_unlock(lock);
        }
        return returnValue;
    }
    atree_unlock(lock);
    return returnValue;
}

static inline
void at_set_insert(void * setParam, void * valuePtr, unsigned int valueSize){
    ATreeSet * set = (ATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    at_set_insert_opt(set,
                      valuePtr,
                      valueSize,
                      set->keyPosition,
                      true,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      true,
                      NULL,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);
}

static inline
void at_set_insert_ops(void * setParam, void * valuePtr, unsigned int valueSize, bool copyIn, void ** replacedWriteBack){
    ATreeSet * set = (ATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    at_set_insert_opt(set,
                      valuePtr,
                      valueSize,
                      set->keyPosition,
                      true,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      copyIn,
                      replacedWriteBack,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);
}



static inline
bool at_set_insert_new(void * setParam, void * valuePtr, unsigned int valueSize){
    ATreeSet * set = (ATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    bool returnValue = at_set_insert_opt(set,
                                         valuePtr, valueSize,
                                         set->keyPosition,
                                         false,
                                         set->extract_key,
                                         set->are_equal,
                                         set->less_than,
                                         true,
                                         NULL,
                                         set->free,
                                         set->freeContext);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}


#ifndef ATREE_SPECIAL_COPY_OUT
#    define ATREE_SPECIAL_COPY_OUT_PARAMS
#    define ATREE_SPECIAL_COPY_OUT_PARAMS_PASS
#    define ATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL
#endif

static inline
void * at_lookup_ValueNode(ValueNode * searchNode,
                           void * key,
                           int keyPosition,
                           void * (*extract_key)(void * v, int keyPos),
                           bool (*are_equal)(void * v1, void * v2),
                           bool (*less_than)(void * v1, void * v2) ATREE_SPECIAL_COPY_OUT_PARAMS){
    ValueNode * currentNode = searchNode;
    while(currentNode != NULL){
        void * nodeKey = extract_key(currentNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->left;
        } else if (are_equal(key, nodeKey)) {
            if(currentNode->deleted){
                return NULL;
            }else{
#ifdef ATREE_SPECIAL_COPY_OUT
                ATREE_SPECIAL_COPY_OUT
#else
                void * toReturn = ATREE_MALLOC(currentNode->valueSize);
                memcpy(toReturn, currentNode->value, currentNode->valueSize);
#endif
                return toReturn;
            }
        } else {
            currentNode = currentNode->right;
        }
    }
    return NULL;
}



static inline
void * at_set_lookup_opt(ATreeSet * set, 
                         void * key,
                         int keyPosition,
                         void * (*extract_key)(void * v, int keyPos),
                         bool (*are_equal)(void * v1, void * v2),
                         bool (*less_than)(void * v1, void * v2) ATREE_SPECIAL_COPY_OUT_PARAMS){
    //Find base node
    BaseOrRouteNode * currentNode = set->root;
    RouteNode * currentRouteNode = NULL;
    while(!currentNode->isBaseNode){
        currentRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentRouteNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentRouteNode->left;
        } else {
            currentNode = currentRouteNode->right;
        }
    }
    //Lock and handle contention
    BaseNode * node = &currentNode->baseOrRoute.base;
    ATreeLock * lock = &node->lock;
    void ** parent = NULL;
    atree_lock(lock);
    if(!node->valid){
        //Retry
        atree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        return at_set_lookup_opt(set,
                                 key,
                                 keyPosition,
                                 extract_key,
                                 are_equal,
                                 less_than ATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    } else if(lock->statistics > ATREE_MAX_CONTENTION_STATISTICS){
        if(currentRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentRouteNode->left == currentNode){
            parent = (void**)&currentRouteNode->left;
        }else{
            parent = (void**)&currentRouteNode->right;
        }
    }
    //Handle empty before and then normal case
    void * returnValue = NULL;
    if(node->root == NULL) {
        returnValue = NULL;
    } else {
        returnValue = at_lookup_ValueNode(node->root,
                                          key,
                                          keyPosition,
                                          extract_key,
                                          are_equal,
                                          less_than ATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    }
    if(parent != NULL){
        if(!contentionReduceSplit(currentNode,
                                  parent,
                                  set,
                                  keyPosition,
                                  extract_key,
                                  less_than)){
            atree_unlock(lock);
        }
        return returnValue;
    }
    atree_unlock(lock);
    return returnValue;
}

static inline
void * at_set_lookup(void * setParam, 
                     void * key){
    ATreeSet * set = (ATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    void * returnValue = at_set_lookup_opt(set,
                                           key,
                                           set->keyPosition,
                                           set->extract_key,
                                           set->are_equal,
                                           set->less_than ATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}

static inline
void * at_set_lookup_ops(void * setParam, 
                         void * key ATREE_SPECIAL_COPY_OUT_PARAMS){
    ATreeSet * set = (ATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    void * returnValue = at_set_lookup_opt(set,
                                           key,
                                           set->keyPosition,
                                           set->extract_key,
                                           set->are_equal,
                                           set->less_than ATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}

static inline
void at_set_delete_ValueNode(ValueNode * searchNode, 
                             BaseNode * baseNode,
                             void * key,
                             int keyPosition,
                             void * (*extract_key)(void * v, int keyPos),
                             bool (*are_equal)(void * v1, void * v2),
                             bool (*less_than)(void * v1, void * v2),
                             void (*custom_free)(void * v, void * context),
                             void * freeContext){
    ValueNode * currentNode = searchNode;
    baseNode->nrOfDeletedElements = baseNode->nrOfDeletedElements + 1;
    while(currentNode != NULL){
        void * nodeKey = extract_key(currentNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->left;
        } else if (are_equal(key, nodeKey)) {
            //Erase
            if(!currentNode->deleted) {
                currentNode->deleted = true;
                if(baseNode->nrOfDeletedElements > (2 *  baseNode->nrOfElements)){
                    baseNode->nrOfElements = countNodes(baseNode->root);
                    baseNode->root = rebuildAt(baseNode->root, baseNode->nrOfElements, baseNode, custom_free, freeContext);
                    baseNode->nrOfDeletedElements = 0;
                }
            }
            return;
        } else {
            currentNode = currentNode->right;
        }
    }
}

static inline
void at_set_delete_opt(ATreeSet * set, 
                       void * key,
                       int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    //Find base node
    BaseOrRouteNode * currentNode = set->root;
    RouteNode * currentRouteNode = NULL;
    while(!currentNode->isBaseNode){
        currentRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentRouteNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentRouteNode->left;
        } else {
            currentNode = currentRouteNode->right;
        }
    }
    //Lock and handle contention
    BaseNode * node = &currentNode->baseOrRoute.base;
    ATreeLock * lock = &node->lock;
    void ** parent = NULL;
    atree_lock(lock);
    if(!node->valid){
        //Retry
        atree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        at_set_delete_opt(set, 
                          key,
                          keyPosition,
                          extract_key,
                          are_equal,
                          less_than,
                          custom_free,
                          freeContext);
        return;
    } else if(lock->statistics > ATREE_MAX_CONTENTION_STATISTICS){
        if(currentRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentRouteNode->left == currentNode){
            parent = (void**)&currentRouteNode->left;
        }else{
            parent = (void**)&currentRouteNode->right;
        }
    }
    //Handle empty before and then normal case
    if(node->root != NULL) {
        at_set_delete_ValueNode(node->root, 
                                node,
                                key,
                                keyPosition,
                                extract_key,
                                are_equal,
                                less_than,
                                custom_free,
                                freeContext);
    }
    if(parent != NULL){
        if(!contentionReduceSplit(currentNode,
                                  parent,
                                  set,
                                  keyPosition,
                                  extract_key,
                                  less_than)){
            atree_unlock(lock);
        }
        return;
    }
    atree_unlock(lock);
}

static inline
void at_set_delete(void * setParam, 
                   void * key,
                   unsigned int keySize){
    (void)keySize;
    ATreeSet * set = (ATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    at_set_delete_opt(set,
                      key,
                      set->keyPosition,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);

}


static inline
void at_set_free_helper_helper(ValueNode * node,
                               void (*custom_free)(void * v, void * context),
                               void * context){
    if(node!=NULL){
        at_set_free_helper_helper(node->left, custom_free, context);
        at_set_free_helper_helper(node->right, custom_free, context);
        custom_free(node->value, context);
        ATREE_FREE(node);
    }
}

static inline
void at_set_free_helper(BaseOrRouteNode * root,
                        void (*custom_free)(void * v, void * context),
                        void * context){
    if(root->isBaseNode){
        at_set_free_helper_helper(root->baseOrRoute.base.root, custom_free, context);
        ATREE_FREE(root);
    }else{
        RouteNode * route = &root->baseOrRoute.route;
        at_set_free_helper(route->left, custom_free, context);
        at_set_free_helper(route->right, custom_free, context);
        custom_free(route->value, context);
        ATREE_FREE(root);
    }
}

static inline
void at_set_destroy(void * setParam){
    ATreeSet * set = (ATreeSet *)setParam;
    at_set_free_helper(set->root, set->free, set->freeContext);
}

static inline
void at_set_free(void * setParam){
    ATreeSet * set = (ATreeSet *)setParam;
    at_set_free_helper(set->root, set->free, set->freeContext);
    ATREE_FREE(set);
}

static inline
unsigned int at_set_size(ATreeSet * set){
    (void)set;
    //TODO Fix code for this!
    return 0;
}


static inline
void at_set_print_helper_helper(ValueNode * node, char * (*to_string)(void * v1)){
    if(node==NULL){
        printf("NULL");
    }else{
        printf("V(");
        char * valueString = to_string(node->value);
        printf("%s", valueString);
        ATREE_FREE(valueString);
        printf(",");
        at_set_print_helper_helper(node->left, to_string);
        printf(",");
        at_set_print_helper_helper(node->right, to_string);
        printf(")");
    }
}

static inline
void at_set_print_helper(BaseOrRouteNode * root, char * (*to_string)(void * v1)){
    if(root->isBaseNode){
        printf("B(");
        at_set_print_helper_helper(root->baseOrRoute.base.root, to_string);
        printf(")");
    }else{
        RouteNode * route = &root->baseOrRoute.route;
        printf("R(");
        char * valueString = to_string(route->value);
        printf("%s", valueString);
        ATREE_FREE(valueString);
        printf(",");
        at_set_print_helper(route->left, to_string);
        printf(",");
        at_set_print_helper(route->right, to_string);
        printf(")");
    }
}

static inline
void at_set_print_helper_helper_dot(ValueNode * node, char * (*to_string)(void * v1)){
    if(node!=NULL){
        if(node->left !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            ATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->left->value);
            printf("%s", valueString);
            ATREE_FREE(valueString);
            printf(";\n");
        }
        if(node->right !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            ATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->right->value);
            printf("%s", valueString);
            ATREE_FREE(valueString);
            printf(";\n");
        }
        at_set_print_helper_helper_dot(node->left, to_string);
        at_set_print_helper_helper_dot(node->right, to_string);
    }
}

static inline
void at_set_print_helper_dot(BaseOrRouteNode * root, char * (*to_string)(void * v1)){
    if(root->isBaseNode){
        printf("B(\n");
        printf("digraph G{\n");
        printf("  graph [ordering=\"out\"];\n");
        at_set_print_helper_helper_dot(root->baseOrRoute.base.root, to_string);
        printf("}\n");
        printf("\n)");
    }else{
        RouteNode * route = &root->baseOrRoute.route;
        printf("R(");
        char * valueString = to_string(route->value);
        printf("%s", valueString);
        ATREE_FREE(valueString);
        printf(",");
        at_set_print_helper_dot(route->left, to_string);
        printf(",");
        at_set_print_helper_dot(route->right, to_string);
        printf(")");
    }
}


static inline
void at_set_print(void * setParam){
    ATreeSet * set = (ATreeSet *)setParam;
    at_set_print_helper(set->root, set->to_string);
    printf("\n\nDOT:\n\n");
    at_set_print_helper_dot(set->root, set->to_string);
    printf("\n\n");
}

static inline
char * at_set_to_string(void * setParam){
    (void)setParam;
    //DODO FIX CODE FOR THIS!
    return NULL;
}

static inline
bool at_set_is_concurrent(){
    return true;
}

static inline
OOSet * oo_at_set_create(unsigned int keyPosition,
                         void * (*extract_key)(void * v, int keyPos),
                         unsigned int (*hash_key)(void * k),
                         bool (*are_equal)(void * v1, void * v2),
                         bool (*less_than)(void * v1, void * v2),
                         char * (*to_string)(void * v1)){
    ATreeSet * set = plain_at_set_create(keyPosition,
                                         extract_key,
                                         hash_key,
                                         are_equal,
                                         less_than,
                                         to_string);
    OOSet * ooset = ATREE_MALLOC(sizeof(OOSet));
    ooset->set = set;
    ooset->m = &ATREE_SET_METHOD_TABLE;
    return ooset;
}


#endif
