#ifndef AVLATREE_SET_H
#define AVLATREE_SET_H

/*
 * The AVL tree related code in this file is derived from the
 * Erlang/OTP ordered_set implementation that was originally written
 * by Patrik Nyblom (https://github.com/bufflig).
 *
 * The AVL tree implementation in Erlang/OTP can be found here:
 * https://github.com/erlang/otp/blob/75937381ec52e1714cfbc5936441b15bb1240bdb/erts/emulator/beam/erl_db_tree.c
 *
 * The following copyright apply to the code derived from Erlang/OTP:
 */

/*
 * %CopyrightBegin%
 *
 * Copyright Ericsson AB 1998-2013. All Rights Reserved.
 *
 * The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved online at http://www.erlang.org/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * %CopyrightEnd%
 */

#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "misc/thread_includes.h"
#include "misc/padded_types.h"
#include "data_structures/oo_set_interface.h"
#include "locks/drmcs_lock.h"
#include "locks/tatas_lock.h"
#include "locks/seq_lock.h"
#ifdef USE_HLE_LOCK
#   include "locks/experimental/hle_lock.h"
#endif

#ifndef AVLATREE_MALLOC
#      define AVLATREE_MALLOC(size) malloc(size)
#endif

#ifndef AVLATREE_FREE
#      define AVLATREE_FREE(data) free(data)
#endif

#ifndef AVLATREE_NEW_VALUE_NODE
#      define AVLATREE_NEW_VALUE_NODE(size, value, context) avlAllocateValueNode(size, value)
#endif

#ifndef AVLATREE_REPLACE_VALUE_NODE
#      define AVLATREE_REPLACE_VALUE_NODE(size, oldValue, value, custom_free, context) avlReplaceValueNode(size,value,oldValue,custom_free,freeContext)
#endif

#ifndef AVLATREE_EXTRACT_KEY_UNTRANSFORMED
#      define AVLATREE_EXTRACT_KEY_UNTRANSFORMED(untransformed) extract_key(valuePtr, keyPosition)
#endif


#define AVLATREE_MAX_CONTENTION_STATISTICS 1000
#define AVLATREE_MIN_CONTENTION_STATISTICS -1000
#define AVLATREE_LOCK_SUCCESS_STATS_CONTRIB -1
#define AVLATREE_LOCK_FAILURE_STATS_CONTRIB 250
#define AVLATREE_REBUILD_CONSTANT 2
#define AVLATREE_STACK_NEED 50
#define AVLATREE_DIR_LEFT 0
#define AVLATREE_DIR_RIGHT 1
#define AVLATREE_DIR_END 2 
#define AVLATREE_FREE_LIST_SIZE 4096
//#define AVLATREE_COUNT_ROUTE_NODES 1

//#define AVLATREE_PTHREAD_LOCK 1
//#define AVLATREE_TATAS_LOCK 1
//#define AVLATREE_SEQ_LOCK 1
#define CATREE_PTHREAD 1
#define CATREE_TATAS 2
#define CATREE_SEQ 3
#define CATREE_HLE 4
#define CATREE_MCS 5
#define CATREE_DRMCS 6

#ifndef CATREE_LOCK_TYPE
#    define AVLATREE_PTHREAD_LOCK 1
#elif CATREE_LOCK_TYPE == CATREE_PTHREAD
#    define AVLATREE_PTHREAD_LOCK 1
#elif CATREE_LOCK_TYPE == CATREE_TATAS
#    define AVLATREE_TATAS_LOCK 1
#elif CATREE_LOCK_TYPE == CATREE_SEQ
#    define AVLATREE_SEQ_LOCK 1
#elif CATREE_LOCK_TYPE == CATREE_HLE
#    define AVLATREE_HLE_LOCK 1
#elif CATREE_LOCK_TYPE == CATREE_MCS
#    define AVLATREE_MCS_LOCK 1
#elif CATREE_LOCK_TYPE == CATREE_DRMCS
#    define AVLATREE_DRMCS_LOCK 1
#endif

#ifdef AVLATREE_PTHREAD_LOCK
#define AVLATREE_LOCK_TYPE pthread_mutex_t
#define AVLATREE_LOCK_INITIALIZER(lock_ptr)                           \
    pthread_mutex_init(lock_ptr, NULL);

    /* pthread_mutexattr_t attr;                                         \ */
    /* pthread_mutexattr_init(&attr);                                    \ */
    /* pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);       \ */
    /* pthread_mutexattr_destroy(&attr); */
#define AVLATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) (!pthread_mutex_trylock(lock_ptr))
#define AVLATREE_LOCK_LOCK_FUNCTION(lock_ptr) pthread_mutex_lock(lock_ptr)
#define AVLATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) pthread_mutex_unlock(lock_ptr)
#define AVLATREE_LOCK_DESTROY_FUNCTION(lock_ptr) pthread_mutex_destroy(lock_ptr)
#endif


#ifdef AVLATREE_TATAS_LOCK
#define AVLATREE_LOCK_TYPE TATASLock
#define AVLATREE_LOCK_INITIALIZER(lock_ptr) tatas_initialize(lock_ptr);
#define AVLATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) tatas_try_lock(lock_ptr)
#define AVLATREE_LOCK_LOCK_FUNCTION(lock_ptr) tatas_lock(lock_ptr)
#define AVLATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) tatas_unlock(lock_ptr)
#define AVLATREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif


#ifdef AVLATREE_SEQ_LOCK
#define AVLATREE_LOCK_TYPE SEQLock
#define AVLATREE_LOCK_INITIALIZER(lock_ptr) seq_initialize(lock_ptr);
#define AVLATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) seq_try_lock(lock_ptr)
#define AVLATREE_LOCK_LOCK_FUNCTION(lock_ptr) seq_lock(lock_ptr)
#define AVLATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) seq_unlock(lock_ptr)
#define AVLATREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif


#ifdef AVLATREE_HLE_LOCK
#define HLE_FAST_READ_PATH 1
#define AVLATREE_LOCK_TYPE HLELock
#define AVLATREE_LOCK_INITIALIZER(lock_ptr) hle_initialize(lock_ptr)
#define AVLATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) hle_try_lock(lock_ptr)
#define AVLATREE_LOCK_LOCK_FUNCTION(lock_ptr) hle_lock(lock_ptr)
#define AVLATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) hle_unlock(lock_ptr)
#define AVLATREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif

#ifdef AVLATREE_MCS_LOCK
#define AVLATREE_LOCK_TYPE MCSLock
#define AVLATREE_LOCK_INITIALIZER(lock_ptr) mcs_initialize(lock_ptr)
#define AVLATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) mcs_try_lock(lock_ptr)
#define AVLATREE_LOCK_LOCK_FUNCTION(lock_ptr) mcs_lock(lock_ptr)
#define AVLATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) mcs_unlock(lock_ptr)
#define AVLATREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif


#ifdef AVLATREE_DRMCS_LOCK
#define AVLATREE_LOCK_TYPE DRMCSLock
#define AVLATREE_LOCK_INITIALIZER(lock_ptr) drmcs_initialize(lock_ptr)
#define AVLATREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) drmcs_try_lock(lock_ptr)
#define AVLATREE_LOCK_LOCK_FUNCTION(lock_ptr) drmcs_lock(lock_ptr)
#define AVLATREE_LOCK_UNLOCK_FUNCTION(lock_ptr) drmcs_unlock(lock_ptr)
#define AVLATREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif

typedef struct {
    AVLATREE_LOCK_TYPE lock;
    // int validation;
    int statistics;
} AvlatreeLock;

typedef struct AVLValueNodeImpl{
    struct AVLValueNodeImpl * left;
    struct AVLValueNodeImpl * right;
    int balance;
    unsigned int valueSize;
    char value[];
} AVLValueNode;

typedef struct {
    struct AVLValueNodeImpl * root;
    AvlatreeLock lock;
    bool valid;
    //    unsigned int nrOfElements;
    //    unsigned int nrOfDeletedElements;
} AVLBaseNode;

typedef struct {
    void * left;
    void * right;
    TATASLock * lock;
    bool valid;
    unsigned int valueSize;
    char value[1];
} AVLRouteNode;

typedef union {
    AVLRouteNode route;
    AVLBaseNode base;
} AVLBaseOrAVLRouteNodeContainer;

typedef struct {
    bool isAVLBaseNode;
    AVLBaseOrAVLRouteNodeContainer baseOrRoute;
} AVLBaseOrAVLRouteNode;

typedef struct {
    DRMCSLock globalLock;
    AVLBaseOrAVLRouteNode * root;
    unsigned int keyPosition;
    void * (*extract_key)(void * v, int keyPos);
    unsigned int (*hash_key)(void * k);
    bool (*are_equal)(void * v1, void * v2);
    bool (*less_than)(void * v1, void * v2);
    char * (*to_string)(void * v1);
    void (*free)(void * v, void * context);
    void * freeContext;
#ifdef AVLATREE_COUNT_ROUTE_NODES
    char pad1[CACHE_LINE_SIZE*2];
    volatile atomic_int nrOfRouteNodes;
#endif
    char pad[CACHE_LINE_SIZE*2];
    TATASLock freeingFreeListLock;
    volatile atomic_intptr_t freeList[AVLATREE_FREE_LIST_SIZE];
    LLPaddedULong freeListIndexCounter;
} AVLATreeSet;

#ifdef CATREE_DELAYED_STATISTICS

typedef struct {
    char pad1[CACHE_LINE_SIZE];
    bool tryLockSucceeded;
    char pad2[2*CACHE_LINE_SIZE - sizeof(bool)];
} LocalTryLockSucessInfo;

extern
_Alignas(CACHE_LINE_SIZE)
__thread LocalTryLockSucessInfo localTryLockInfo = {.tryLockSucceeded = false};

#endif

extern
_Alignas(CACHE_LINE_SIZE)
OOSetMethodTable AVLATREE_SET_METHOD_TABLE;

static inline
void avlatree_lock_init(AvlatreeLock * lock){
    lock->statistics = 0;
    //    lock->validation = 0;

    AVLATREE_LOCK_INITIALIZER(&lock->lock);
}

static inline
void avlatree_lock(AvlatreeLock * lock){
    if(AVLATREE_LOCK_TRY_LOCK_FUNCTION(&lock->lock)){
        lock->statistics += AVLATREE_LOCK_SUCCESS_STATS_CONTRIB;
        return;
    }
    AVLATREE_LOCK_LOCK_FUNCTION(&lock->lock);
    lock->statistics += AVLATREE_LOCK_FAILURE_STATS_CONTRIB;
}


static inline
bool avlatree_trylock(AvlatreeLock * lock){
    if(AVLATREE_LOCK_TRY_LOCK_FUNCTION(&lock->lock)){
        /* lock->validation++; */
        /* assert(lock->validation == 1); */
        return true;
    }else{
        return false;
    }
}


static inline
void avlatree_unlock(AvlatreeLock * lock){
    /* lock->validation--; */
    /* assert(lock->validation == 0); */
    AVLATREE_LOCK_UNLOCK_FUNCTION(&lock->lock);
}

static inline
AVLBaseOrAVLRouteNode * allocateAVLBaseNode(){
    AVLBaseOrAVLRouteNode * newAVLBase = AVLATREE_MALLOC(sizeof(AVLBaseOrAVLRouteNode));
    newAVLBase->isAVLBaseNode = true;
    avlatree_lock_init(&newAVLBase->baseOrRoute.base.lock);
    newAVLBase->baseOrRoute.base.root = NULL;
    //    newAVLBase->baseOrRoute.base.nrOfElements = 0;
    //    newAVLBase->baseOrRoute.base.nrOfDeletedElements = 0;
    newAVLBase->baseOrRoute.base.valid = true;
    return newAVLBase;
}

static inline
AVLBaseOrAVLRouteNode * avlAllocateRouteNode(unsigned int valueSize,
                                             void * value,
                                             AVLBaseOrAVLRouteNode * newAVLBaseLeft,
                                             AVLBaseOrAVLRouteNode * newAVLBaseRight){
    AVLBaseOrAVLRouteNode * newRoute = 
        AVLATREE_MALLOC(sizeof(AVLBaseOrAVLRouteNode) + valueSize - sizeof(char));
    newRoute->isAVLBaseNode = false;
    newRoute->baseOrRoute.route.lock = plain_tatas_create();
    newRoute->baseOrRoute.route.valid = true;
    newRoute->baseOrRoute.route.left = newAVLBaseLeft;
    newRoute->baseOrRoute.route.right = newAVLBaseRight;
    newRoute->baseOrRoute.route.valueSize = valueSize;
    memcpy(newRoute->baseOrRoute.route.value, value, valueSize);
    return newRoute;
}



static inline
AVLValueNode * avlAllocateValueNode(unsigned int size, void * value){
    AVLValueNode * newAVLValueNode = AVLATREE_MALLOC(sizeof(AVLValueNode) + size);
    newAVLValueNode->left = NULL;
    newAVLValueNode->right = NULL;
    newAVLValueNode->valueSize = size;
    newAVLValueNode->balance = 0;
    memcpy(newAVLValueNode->value, value, size);
    return newAVLValueNode;
}

static inline
AVLValueNode * avlReplaceValueNode(unsigned int size,
                                   void * value,
                                   AVLValueNode * toReplace,
                                   void (*custom_free)(void * v, void * context),
                                   void * freeContext){
    if(toReplace->valueSize < size){
        AVLValueNode * toReturn = avlAllocateValueNode(size, value);
        custom_free(toReplace, freeContext);
        return toReturn;
    }else{
        toReplace->left = NULL;
        toReplace->right = NULL;
        toReplace->valueSize = size;
        toReplace->balance = 0;
        memcpy(toReplace->value, value, size);
        return toReplace;
    }

}

  /* private var root:Node = AVLBaseNode(null) */

  /* def printRootStructure() = printStructure(root) */

  /* def printStructure(currentNode:Node):String = currentNode match { */
  /*   case AVLBaseNode(root, _, _, nrOfElements) => "B("+ (if(root == null) null else(root.key)) + ","+nrOfElements+")" */
  /*   case AVLRouteNode(key, left, right) => "R("+key+","+printStructure(left)+","+printStructure(right)+")" */
  /* } */


static void avlEnqueueInFreeList(AVLATreeSet * set, AVLBaseOrAVLRouteNode * toEnqueue){
    while(true){
        unsigned long slot = atomic_fetch_add(&set->freeListIndexCounter.value, 1);
        if(slot == AVLATREE_FREE_LIST_SIZE){
            // We are responsible for doing the freeing
            // Ensure that no one still is accessing old data
            AVLBaseOrAVLRouteNode * elementsToFree[AVLATREE_FREE_LIST_SIZE];
            drmcs_runlock(&set->globalLock);
            drmcs_lock(&set->globalLock);
            drmcs_unlock(&set->globalLock);
            drmcs_rlock(&set->globalLock);
            for(int i = 0; i < AVLATREE_FREE_LIST_SIZE; i++){
                AVLBaseOrAVLRouteNode * elementToFree;
                while((elementToFree = (AVLBaseOrAVLRouteNode *)atomic_load(&set->freeList[i])) == NULL){
                    thread_yield();//This should not happen
                }
                elementsToFree[i] = elementToFree;
                atomic_store_explicit(&set->freeList[i], (intptr_t)NULL, memory_order_relaxed);
            }
            //Enqueue our element
            atomic_store_explicit(&set->freeList[0], (intptr_t)toEnqueue, memory_order_relaxed);
            //Let other threads start to enqueue
            atomic_store(&set->freeListIndexCounter.value, 1);
            for(int i = 0; i < AVLATREE_FREE_LIST_SIZE; i++){
                AVLBaseOrAVLRouteNode * elementToFree = elementsToFree[i];
                if(elementToFree->isAVLBaseNode){
                    AVLATREE_LOCK_DESTROY_FUNCTION(&elementToFree->baseOrRoute.base.lock.lock); //Free deleted base
                    AVLATREE_FREE(elementToFree);
                }else{
                    free(elementToFree->baseOrRoute.route.lock); //Free deleted route
                    AVLATREE_FREE(elementToFree);   
                }

            }
            return;
        }else if (slot > AVLATREE_FREE_LIST_SIZE){
            //We need to wait then retry
            drmcs_runlock(&set->globalLock);
            while(atomic_load(&set->freeListIndexCounter.value) > AVLATREE_FREE_LIST_SIZE){
                thread_yield();
            }
            drmcs_rlock(&set->globalLock);
        } else {
            // Just write our node to our slot
            atomic_store_explicit(&set->freeList[slot], (intptr_t)toEnqueue, memory_order_relaxed);
            return;
        }
    }
}


//http://www.hackersdelight.org/hdcodetxt/nlz.c.txt
static inline
unsigned int avlNlz1(unsigned int x) {
   int n;

   if (x == 0) return(32);
   n = 0;
   if (x <= 0x0000FFFF) {n = n +16; x = x <<16;}
   if (x <= 0x00FFFFFF) {n = n + 8; x = x << 8;}
   if (x <= 0x0FFFFFFF) {n = n + 4; x = x << 4;}
   if (x <= 0x3FFFFFFF) {n = n + 2; x = x << 2;}
   if (x <= 0x7FFFFFFF) {n = n + 1;}
   return n;
}

static inline
unsigned int avlLog2nlz(unsigned int bits) {
    if( bits == 0 ) {
        return 0; 
    }else {
        return 31 - avlNlz1( bits );
    }
}

 static inline void aat_standard_free(void * toFree, void * freeContext){
     (void)freeContext;
     AVLATREE_FREE(toFree);
 }

static inline
void plain_aat_set_init(AVLATreeSet * set,
                       unsigned int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       unsigned int (*hash_key)(void * k),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       char * (*to_string)(void * v1),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    drmcs_initialize(&set->globalLock);
    set->keyPosition = keyPosition;
    set->extract_key = extract_key;
    set->hash_key = hash_key;
    set->are_equal = are_equal;
    set->less_than = less_than;
    set->to_string = to_string;
    if(custom_free == NULL){
        set->free = aat_standard_free;
        set->freeContext = NULL;
    }else{
        set->free = custom_free;
        set->freeContext = freeContext;
    }
    set->root = allocateAVLBaseNode();
    tatas_initialize(&set->freeingFreeListLock);
    for(int i = 0; i < AVLATREE_FREE_LIST_SIZE; i++){
        atomic_init(&set->freeList[i], (intptr_t)NULL);
    }
    atomic_init(&set->freeListIndexCounter.value, 0);
#ifdef AVLATREE_COUNT_ROUTE_NODES
    atomic_init(&set->nrOfRouteNodes, 0);
#endif
}


static inline
AVLATreeSet * plain_aat_set_create(unsigned int keyPosition,
                               void * (*extract_key)(void * v, int keyPos),
                               unsigned int (*hash_key)(void * k),
                               bool (*are_equal)(void * v1, void * v2),
                               bool (*less_than)(void * v1, void * v2),
                               char * (*to_string)(void * v1)){
    AVLATreeSet * set = AVLATREE_MALLOC(sizeof(AVLATreeSet));
    plain_aat_set_init(set,
                      keyPosition,
                      extract_key,
                      hash_key,
                      are_equal,
                      less_than,
                      to_string,
                      NULL,
                      NULL);
    return set;
}

static inline
int avl_compare_hash_codes2(unsigned int code1, unsigned int code2){
    return code1 - code2;
}

static inline
unsigned int avlCountNodes(AVLValueNode * node){ 
    if(node == NULL) return 0;
    else if(node->left == NULL ) return avlCountNodes(node->right) + 1;
    else if(node->right == NULL ) return avlCountNodes(node->left) + 1;
    else return avlCountNodes(node->left) + 1 + avlCountNodes(node->right);
}


static inline
AVLValueNode * avlInsertAVLValueNode(AVLValueNode * insertTo,
                                     AVLValueNode * insertValueNode,
                                     int keyPosition,
                                     void * (*extract_key)(void * v, int keyPos),
                                     bool (*less_than)(void * v1, void * v2)){
    /* Non recursive insertion in AVL tree, building our own stack */
    AVLValueNode **tstack[AVLATREE_STACK_NEED];
    int tpos = 0;
    int dstack[AVLATREE_STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    AVLValueNode * base = insertTo;
    AVLValueNode **this = &base;
    void * key;
    int dir;
    AVLValueNode *p1, *p2, *p;

    key = extract_key(insertValueNode->value, keyPosition);

    dstack[dpos++] = AVLATREE_DIR_END;
    for (;;)
	if (!*this) { /* Found our place */
	    state = 1;
	    *this = insertValueNode;
	    (*this)->balance = 0;
	    (*this)->left = (*this)->right = NULL;
	    break;
	} else if (less_than(key, extract_key((*this)->value, keyPosition))) {
	    /* go lefts */
	    dstack[dpos++] = AVLATREE_DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else { /* go right */
	    dstack[dpos++] = AVLATREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	}

    while (state && ( dir = dstack[--dpos] ) != AVLATREE_DIR_END) {
	this = tstack[--tpos];
	p = *this;
	if (dir == AVLATREE_DIR_LEFT) {
	    switch (p->balance) {
	    case 1:
		p->balance = 0;
		state = 0;
		break;
	    case 0:
		p->balance = -1;
		break;
	    case -1: /* The icky case */
		p1 = p->left;
		if (p1->balance == -1) { /* Single LL rotation */
		    p->left = p1->right;
		    p1->right = p;
		    p->balance = 0;
		    (*this) = p1;
		} else { /* Double RR rotation */
		    p2 = p1->right;
		    p1->right = p2->left;
		    p2->left = p1;
		    p->left = p2->right;
		    p2->right = p;
		    p->balance = (p2->balance == -1) ? +1 : 0;
		    p1->balance = (p2->balance == 1) ? -1 : 0;
		    (*this) = p2;
		}
		(*this)->balance = 0;
		state = 0;
		break;
	    }
	} else { /* dir == DIR_RIGHT */
	    switch (p->balance) {
	    case -1:
		p->balance = 0;
		state = 0;
		break;
	    case 0:
		p->balance = 1;
		break;
	    case 1:
		p1 = p->right;
		if (p1->balance == 1) { /* Single RR rotation */
		    p->right = p1->left;
		    p1->left = p;
		    p->balance = 0;
		    (*this) = p1;
		} else { /* Double RL rotation */
		    p2 = p1->left;
		    p1->left = p2->right;
		    p2->right = p1;
		    p->right = p2->left;
		    p2->left = p;
		    p->balance = (p2->balance == 1) ? -1 : 0;
		    p1->balance = (p2->balance == -1) ? 1 : 0;
		    (*this) = p2;
		}
		(*this)->balance = 0; 
		state = 0;
		break;
	    }
	}
    }
    return base;
}

static inline int avlValidate(AVLValueNode * toTest){
    (void)toTest;
    if(toTest == NULL){
        return 0;
    }else{
        int hl = avlValidate(toTest->left);
        int hr = avlValidate(toTest->right);
        if(toTest->balance == 0 && hl != hr){
            printf("FAIL 1 %d %d\n", hl, hr);
            assert(0);
        }else if(toTest->balance == -1 && (hr - hl) != -1){
            printf("FAIL 2\n");
            assert(0);
        }else if(toTest->balance == 1 && (hr - hl) != 1){
            printf("FAIL 3\n");
            assert(0);
        }else if(toTest->balance > 1 || toTest->balance < -1){
            printf("FAIL 4\n");
            assert(0);
        }
        if(hl > hr){
            return hl + 1;
        }else{
            return hr + 1;
        }
    }
    return 0;
}

static inline
bool avlContentionReduceSplit(AVLBaseOrAVLRouteNode *  baseNodeContainer,
                              void ** parent,
                              AVLATreeSet * set,
                              int keyPosition,
                              void * (*extract_key)(void * v, int keyPos),
                              bool (*less_than)(void * v1, void * v2)) {
    (void)less_than;
    (void)extract_key;
    (void)keyPosition;
    AVLBaseNode * node = &baseNodeContainer->baseOrRoute.base;
    if(node == NULL || node->root == NULL || (node->root->left == NULL && node->root->right == NULL)){
        node->lock.statistics = 0;
        return false;
    }
    AVLValueNode * splitAVLValueNode = NULL;
    if (node->root->left == NULL) {
          //To get non empty split
          AVLValueNode * insertTo = node->root->right;
          node->root->right = NULL;
          splitAVLValueNode = avlInsertAVLValueNode(insertTo,
                                                    node->root,
                                                    keyPosition,
                                                    extract_key,
                                                    less_than);
      }else{
          splitAVLValueNode = node->root;
      }

      void * splitKey = splitAVLValueNode->value;
      unsigned int splitKeyValueSize = splitAVLValueNode->valueSize;
      AVLValueNode * leftAVLBaseRoot = splitAVLValueNode->left;
      splitAVLValueNode->left = NULL;
      AVLValueNode * rightAVLBaseRoot = NULL;
      if(splitAVLValueNode->right == NULL){
          rightAVLBaseRoot = splitAVLValueNode;
          splitAVLValueNode->balance = 0;
      } else {
          //To narrow down faster
          AVLValueNode * base = splitAVLValueNode->right;
          splitAVLValueNode->right = NULL;
          rightAVLBaseRoot = avlInsertAVLValueNode(base,
                                                   splitAVLValueNode,
                                                   keyPosition,
                                                   extract_key,
                                                   less_than);
      }
      //We can not use the same because it might contain less than what a thread waiting for expect
      AVLBaseOrAVLRouteNode * newAVLBaseRight = allocateAVLBaseNode();
      AVLBaseNode * rightBCont = &newAVLBaseRight->baseOrRoute.base;
      rightBCont->root = rightAVLBaseRoot;

      AVLBaseOrAVLRouteNode * newAVLBaseLeft = allocateAVLBaseNode();
      AVLBaseNode * leftBCont = &newAVLBaseLeft->baseOrRoute.base;
      leftBCont->root = leftAVLBaseRoot;
      
      AVLBaseOrAVLRouteNode * newRoute = avlAllocateRouteNode(splitKeyValueSize,
                                                              splitKey,
                                                              newAVLBaseLeft,
                                                              newAVLBaseRight);
      //Link in new route
      *parent = newRoute;
      //Unlock threads waiting for old node
      node->valid = false;
      avlatree_unlock(&node->lock);
      //Done!
      avlEnqueueInFreeList(set, baseNodeContainer);
#ifdef AVLATREE_COUNT_ROUTE_NODES
      int count = atomic_fetch_add(&set->nrOfRouteNodes, 1);
      printf("%d\n", count);
#endif
      return true;
}


static inline
int avlComputeHeight(AVLValueNode * valueNode){
    if(valueNode == NULL){
        return 0;
    } else {
        AVLValueNode * currentNode = valueNode;
        int hightSoFar = 1;
        while(currentNode->left != NULL || currentNode->right != NULL){
            if(currentNode->balance == -1){
                currentNode = currentNode->left;
            }else{
                currentNode = currentNode->right;
            }
            hightSoFar = hightSoFar + 1;
        }
        return hightSoFar;
    }
}

static inline
void * avlFindMinValue(AVLValueNode * root, int * valueSizeWriteBack, void * freeContext){
    (void)freeContext;
    AVLValueNode * currentNode = root;
    while(currentNode->left != NULL){
        currentNode = currentNode->left;
    }
    *valueSizeWriteBack = currentNode->valueSize;
    void * copiedValue = AVLATREE_MALLOC(currentNode->valueSize);
    memcpy(copiedValue, currentNode->value, currentNode->valueSize);
    return copiedValue;
}


static inline
void * avlFindMaxValue(AVLValueNode * root, int * valueSizeWriteBack, void * freeContext){
    (void)freeContext;
    AVLValueNode * currentNode = root;
    while(currentNode->right != NULL){
        currentNode = currentNode->right;
    }
    *valueSizeWriteBack = currentNode->valueSize;
    void * copiedValue = AVLATREE_MALLOC(currentNode->valueSize);
    memcpy(copiedValue, currentNode->value, currentNode->valueSize);
    return copiedValue;
}


static inline
AVLValueNode * aat_set_delete_AVLValueNode(AVLBaseNode * baseNode,
                                           void * key,
                                           int keyPosition,
                                           void * (*extract_key)(void * v, int keyPos),
                                           bool (*are_equal)(void * v1, void * v2),
                                           bool (*less_than)(void * v1, void * v2),
                                           void (*custom_free)(void * v, void * context),
                                           void * freeContext,
                                           bool returnDeleted);

static inline
AVLValueNode * avlJoinTrees(AVLBaseNode * leftBase,
                            AVLBaseNode * rightBase,
                            int keyPosition,
                            void * (*extract_key)(void * v, int keyPos),
                            bool (*are_equal)(void * v1, void * v2),
                            bool (*less_than)(void * v1, void * v2),
                            void (*custom_free)(void * v, void * context),
                            void * freeContext){
    AVLValueNode **tstack[AVLATREE_STACK_NEED];
    int tpos = 0;
    int dstack[AVLATREE_STACK_NEED+1];
    int dpos = 0;
    int state = 1;
    AVLValueNode **this;
    int dir;
    AVLValueNode *p1, *p2, *p;
    dstack[dpos++] = AVLATREE_DIR_END;

    int leftHeight = avlComputeHeight(leftBase->root);
    int rightHeight = avlComputeHeight(rightBase->root);
    if(leftHeight >= rightHeight){
        int rightMinValueSize;
        void * rightMinValue = avlFindMinValue(rightBase->root, &rightMinValueSize, freeContext);
        AVLValueNode * newRoot = 
            aat_set_delete_AVLValueNode(rightBase,
                                        extract_key(rightMinValue, keyPosition),
                                        keyPosition,
                                        extract_key,
                                        are_equal,
                                        less_than,
                                        custom_free,
                                        freeContext,
                                        true);
        AVLATREE_FREE(rightMinValue);
        int newRightHeight = avlComputeHeight(rightBase->root);
        // Find a node v on the rightmost path from the root of T1 , whose height is either h or h + 1, as follows:
        // From: http://www.cs.toronto.edu/~avner/teaching/263/A/2sol.pdf
        // v ← root(T1 )
        // h' ← h1
        // while h > h + 1 do
        //    if balance factor (v) = -1
        //    then h' ← h' − 2
        //    else h' ← h' − 1
        //    v ← rightchild(v)
        //AVLValueNode * prevNode = NULL;
        AVLValueNode * currentNode = leftBase->root;
        this = &leftBase->root;
        int currentHeight = leftHeight;
        while(currentHeight > newRightHeight + 1){
            if(currentNode->balance == -1){
                currentHeight = currentHeight - 2;
            }else{
                currentHeight = currentHeight - 1;                
            }
            //prevNode = currentNode;
	    dstack[dpos++] = AVLATREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
            currentNode = currentNode->right;
        }
        newRoot->left = currentNode;
        newRoot->right = rightBase->root;
        newRoot->balance = newRightHeight - currentHeight;
        *this = newRoot;
    }else{
        //This case is symetric to the previous case
        int leftMaxValueSize;//f
        void * leftMaxValue = avlFindMaxValue(leftBase->root, &leftMaxValueSize, freeContext);//f
        AVLValueNode * newRoot = 
            aat_set_delete_AVLValueNode(leftBase,
                                        extract_key(leftMaxValue, keyPosition),//f
                                        keyPosition,//f
                                        extract_key,//f
                                        are_equal,//f
                                        less_than,//f
                                        custom_free,//f
                                        freeContext,
                                        true);//f
        AVLATREE_FREE(leftMaxValue);//f
        int newLeftHeight = avlComputeHeight(leftBase->root);//f
        // Find a node v on the rightmost path from the root of T1 , whose height is either h or h + 1, as follows:
        // From: http://www.cs.toronto.edu/~avner/teaching/263/A/2sol.pdf
        // v ← root(T1 )
        // h' ← h1
        // while h > h + 1 do
        //    if balance factor (v) = -1
        //    then h' ← h' − 2
        //    else h' ← h' − 1
        //    v ← rightchild(v)
        // AVLValueNode * prevNode = NULL;//f
        AVLValueNode * currentNode = rightBase->root;//f
        this = &rightBase->root;//f
        int currentHeight = rightHeight;//f
        while(currentHeight > newLeftHeight + 1){//f
            if(currentNode->balance == 1){//f
                currentHeight = currentHeight - 2;//f
            }else{
                currentHeight = currentHeight - 1;//f             
            }
            //prevNode = currentNode;//f
	    dstack[dpos++] = AVLATREE_DIR_LEFT;//f
	    tstack[tpos++] = this;//f
	    this = &((*this)->left);//f
            currentNode = currentNode->left;//f
        }
        newRoot->right = currentNode;//f
        newRoot->left = leftBase->root;//f
        newRoot->balance = currentHeight - newLeftHeight;//f
        *this = newRoot;//f
    }
    //Now we need to continue as if this was during the insert 
    while (state && ( dir = dstack[--dpos] ) != AVLATREE_DIR_END) {
	this = tstack[--tpos];
	p = *this;
	if (dir == AVLATREE_DIR_LEFT) {
	    switch (p->balance) {
	    case 1:
		p->balance = 0;
		state = 0;
		break;
	    case 0:
		p->balance = -1;
		break;
	    case -1: /* The icky case */
		p1 = p->left;
		if (p1->balance == -1) { /* Single LL rotation */
		    p->left = p1->right;
		    p1->right = p;
		    p->balance = 0;
		    (*this) = p1;
		} else { /* Double RR rotation */
		    p2 = p1->right;
		    p1->right = p2->left;
		    p2->left = p1;
		    p->left = p2->right;
		    p2->right = p;
		    p->balance = (p2->balance == -1) ? +1 : 0;
		    p1->balance = (p2->balance == 1) ? -1 : 0;
		    (*this) = p2;
		}
		(*this)->balance = 0;
		state = 0;
		break;
	    }
	} else { /* dir == DIR_RIGHT */
	    switch (p->balance) {
	    case -1:
		p->balance = 0;
		state = 0;
		break;
	    case 0:
		p->balance = 1;
		break;
	    case 1:
		p1 = p->right;
		if (p1->balance == 1) { /* Single RR rotation */
		    p->right = p1->left;
		    p1->left = p;
		    p->balance = 0;
		    (*this) = p1;
		} else { /* Double RL rotation */
		    p2 = p1->left;
		    p1->left = p2->right;
		    p2->right = p1;
		    p->right = p2->left;
		    p2->left = p;
		    p->balance = (p2->balance == 1) ? -1 : 0;
		    p1->balance = (p2->balance == -1) ? 1 : 0;
		    (*this) = p2;
		}
		(*this)->balance = 0; 
		state = 0;
		break;
	    }
	}
    }
    // Return the joined tree!
    if(leftHeight >= rightHeight){
        return leftBase->root;
    }else{
        return rightBase->root;
    }
}

static inline
bool aat_update_helper_AVLValueNode(AVLBaseNode * baseNode,
                                    void * valuePtr,
                                    unsigned int valueSize,
                                    int keyPosition,
                                    bool overwrite,
                                    void * (*extract_key)(void * v, int keyPos),
                                    bool (*are_equal)(void * v1, void * v2),
                                    bool (*less_than)(void * v1, void * v2),
                                    void ** replacedWriteBack,
                                    void (*custom_free)(void * v, void * context),
                                    void * freeContext){
    (void)replacedWriteBack;
    (void)custom_free;
    (void)freeContext;
    /* Non recursive insertion in AVL tree, building our own stack */
    AVLValueNode **tstack[AVLATREE_STACK_NEED];
    int tpos = 0;
    int dstack[AVLATREE_STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    AVLValueNode **this = &baseNode->root;
    void * key;
    int dir;
    AVLValueNode *p1, *p2, *p;

    key = AVLATREE_EXTRACT_KEY_UNTRANSFORMED(valuePtr);

    dstack[dpos++] = AVLATREE_DIR_END;
    for (;;)
	if (!*this) { /* Found our place */
	    state = 1;
            *this = AVLATREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
	    (*this)->balance = 0;
	    (*this)->left = (*this)->right = NULL;
	    break;
	} else if (less_than(key, extract_key((*this)->value, keyPosition))) {
	    /* go lefts */
	    dstack[dpos++] = AVLATREE_DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else if (are_equal(key, extract_key((*this)->value, keyPosition))) { 
            if(overwrite){
                AVLValueNode * old = *this;
                AVLValueNode * oldLeft = old->left;
                AVLValueNode * oldRight = old->right;
                int oldBalance = old->balance;
                *this = AVLATREE_REPLACE_VALUE_NODE(valueSize, old, valuePtr, custom_free, freeContext);
                (*this)->left = oldLeft;
                (*this)->right = oldRight;
                (*this)->balance = oldBalance;
            }else{
                return false;
            }
	    break;
	} else { /* go right */
	    dstack[dpos++] = AVLATREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	}

    while (state && ( dir = dstack[--dpos] ) != AVLATREE_DIR_END) {
        this = tstack[--tpos];
        p = *this;
        if (dir == AVLATREE_DIR_LEFT) {
            switch (p->balance) {
            case 1:
        	p->balance = 0;
        	state = 0;
        	break;
            case 0:
        	p->balance = -1;
        	break;
            case -1: /* The icky case */
        	p1 = p->left;
        	if (p1->balance == -1) { /* Single LL rotation */
        	    p->left = p1->right;
        	    p1->right = p;
        	    p->balance = 0;
        	    (*this) = p1;
        	} else { /* Double RR rotation */
        	    p2 = p1->right;
        	    p1->right = p2->left;
        	    p2->left = p1;
        	    p->left = p2->right;
        	    p2->right = p;
        	    p->balance = (p2->balance == -1) ? +1 : 0;
        	    p1->balance = (p2->balance == 1) ? -1 : 0;
        	    (*this) = p2;
        	}
        	(*this)->balance = 0;
        	state = 0;
        	break;
            }
        } else { /* dir == DIR_RIGHT */
            switch (p->balance) {
            case -1:
        	p->balance = 0;
        	state = 0;
        	break;
            case 0:
        	p->balance = 1;
        	break;
            case 1:
        	p1 = p->right;
        	if (p1->balance == 1) { /* Single RR rotation */
        	    p->right = p1->left;
        	    p1->left = p;
        	    p->balance = 0;
        	    (*this) = p1;
        	} else { /* Double RL rotation */
        	    p2 = p1->left;
        	    p1->left = p2->right;
        	    p2->right = p1;
        	    p->right = p2->left;
        	    p2->left = p;
        	    p->balance = (p2->balance == 1) ? -1 : 0;
        	    p1->balance = (p2->balance == -1) ? 1 : 0;
        	    (*this) = p2;
        	}
        	(*this)->balance = 0;
        	state = 0;
        	break;
            }
        }
    }
    return true;
}


static inline AVLBaseOrAVLRouteNode * avlFindLeftmostBaseNode(AVLBaseOrAVLRouteNode *  baseOrRouteParam,
                                                              AVLBaseOrAVLRouteNode ** parentRouteWriteBack){
    AVLBaseOrAVLRouteNode *  baseOrRoute = baseOrRouteParam; 
    AVLBaseOrAVLRouteNode *  prevNode = NULL;
    while(!baseOrRoute->isAVLBaseNode){
        prevNode = baseOrRoute;
        baseOrRoute = baseOrRoute->baseOrRoute.route.left;
    }
    *parentRouteWriteBack = prevNode;
    return baseOrRoute;
}

static inline AVLBaseOrAVLRouteNode * avlFindRightmostBaseNode(AVLBaseOrAVLRouteNode *  baseOrRouteParam,
                                                               AVLBaseOrAVLRouteNode ** parentRouteWriteBack){
    AVLBaseOrAVLRouteNode *  baseOrRoute = baseOrRouteParam; 
    AVLBaseOrAVLRouteNode *  prevNode = NULL;
    while(!baseOrRoute->isAVLBaseNode){
        prevNode = baseOrRoute;
        baseOrRoute = baseOrRoute->baseOrRoute.route.right;
    }
    *parentRouteWriteBack = prevNode;
    return baseOrRoute;
}

static inline AVLBaseOrAVLRouteNode * avlFindParentOf(AVLBaseOrAVLRouteNode * container,
                                                       void * value,
                                                       AVLATreeSet * set,
                                                       int keyPosition,
                                                       void * (*extract_key)(void * v, int keyPos),
                                                       bool (*less_than)(void * v1, void * v2)){
    if(set->root == container){
        return NULL;
    }
    AVLBaseOrAVLRouteNode * currentNode = set->root;
    AVLBaseOrAVLRouteNode * prevNode = NULL;
    void * key = extract_key(value, keyPosition);
    while(currentNode != container){
        void * nodeKey = extract_key(currentNode->baseOrRoute.route.value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->baseOrRoute.route.left;
        } else {
            currentNode = currentNode->baseOrRoute.route.right;
        }
    }
    return prevNode;
}


static inline
bool avlNoContentionMerge(AVLBaseOrAVLRouteNode *  baseNodeContainer,
                          AVLBaseOrAVLRouteNode * parentRoute,
                          AVLATreeSet * set,
                          int keyPosition,
                          void * (*extract_key)(void * v, int keyPos),
                          bool (*are_equal)(void * v1, void * v2),
                          bool (*less_than)(void * v1, void * v2),
                          void (*custom_free)(void * v, void * context),
                          void * freeContext) {
    AVLBaseNode * node = &baseNodeContainer->baseOrRoute.base;
    if(parentRoute == set->root){
        node->lock.statistics = 0;
        return false;
    }
    AVLRouteNode * parent = &parentRoute->baseOrRoute.route;
    AVLBaseNode * neighbourBase;
    bool neighbourLeft = true;
    AVLBaseOrAVLRouteNode * neighbourBaseContainer;
    AVLBaseOrAVLRouteNode * neighbourBaseParentRoute = NULL;
    if(baseNodeContainer == parent->left){
        neighbourLeft = false;
        neighbourBaseContainer = avlFindLeftmostBaseNode(parent->right, &neighbourBaseParentRoute);
    } else {
        neighbourBaseContainer = avlFindRightmostBaseNode(parent->left, &neighbourBaseParentRoute);
    }
    if(neighbourBaseParentRoute == NULL){
        neighbourBaseParentRoute = parentRoute;
    }
    neighbourBase = &neighbourBaseContainer->baseOrRoute.base;
    if(!avlatree_trylock(&neighbourBase->lock)){
        node->lock.statistics = 0;
        return false; // Lets retry later
    }
    
    if(!neighbourBase->valid){
        avlatree_unlock(&neighbourBase->lock);
        node->lock.statistics = 0;
        return false; // Lets retry later
    }
    // Ready to do the merge
    AVLBaseOrAVLRouteNode * newNeighbourBaseContainer = allocateAVLBaseNode();
    AVLBaseNode * newNeighbourBase = &newNeighbourBaseContainer->baseOrRoute.base;
    if(neighbourBase->root == NULL){
        newNeighbourBase->root = node->root;
    }else if(node->root != NULL){
        AVLBaseNode * leftBase;
        AVLBaseNode * rightBase;
        if(neighbourLeft){
            leftBase = neighbourBase;
            rightBase = node;
        }else{
            leftBase = node;
            rightBase = neighbourBase;
        }
        newNeighbourBase->root = avlJoinTrees(leftBase,
                                              rightBase,
                                              keyPosition,
                                              extract_key,
                                              are_equal,
                                              less_than,
                                              custom_free,
                                              freeContext);
    }
    //Take out the node
    //Lock parent
    tatas_lock(parent->lock);
    //Find parent of parent and lock if posible
    AVLBaseOrAVLRouteNode * parentOfParentRoute = NULL;
    do{
        if(parentOfParentRoute != NULL){
            tatas_unlock(parentOfParentRoute->baseOrRoute.route.lock);
        }
        parentOfParentRoute = 
            avlFindParentOf(parentRoute,
                         parent->value,
                         set,
                         keyPosition,
                         extract_key,
                         less_than);
        if(parentOfParentRoute == NULL){
            //Parent of parent is root, we are safe becaue it can't be deleted
            break;
        }
        tatas_lock(parentOfParentRoute->baseOrRoute.route.lock);
    }while(!parentOfParentRoute->baseOrRoute.route.valid);
    //Parent and parent of parent is safe
    //We can now unlink the parent route 
    AVLBaseOrAVLRouteNode ** parentInLinkPtrPtr = NULL;
    if(parentOfParentRoute == NULL){
        parentInLinkPtrPtr = &set->root;
    }else if(parentOfParentRoute->baseOrRoute.route.left == parentRoute){
        parentInLinkPtrPtr = (AVLBaseOrAVLRouteNode **)&parentOfParentRoute->baseOrRoute.route.left;
    }else{
        parentInLinkPtrPtr = (AVLBaseOrAVLRouteNode **)&parentOfParentRoute->baseOrRoute.route.right;
    }
    if(neighbourLeft){
        *parentInLinkPtrPtr = parent->left;
    }else{
        *parentInLinkPtrPtr = parent->right;
    }
    //Unlink should have happened YEY
    //Unlock the locks
    parent->valid = false;
    tatas_unlock(parent->lock);
    if(parentOfParentRoute != NULL){
        tatas_unlock(parentOfParentRoute->baseOrRoute.route.lock);
    }
    //Link in new neighbour base
    if(*parentInLinkPtrPtr == neighbourBaseContainer){
        *parentInLinkPtrPtr = newNeighbourBaseContainer;
    }else if(neighbourBaseParentRoute->baseOrRoute.route.left == neighbourBaseContainer){
        neighbourBaseParentRoute->baseOrRoute.route.left = newNeighbourBaseContainer;
    }else{
        neighbourBaseParentRoute->baseOrRoute.route.right = newNeighbourBaseContainer;
    }
    neighbourBase->valid = false;
    avlatree_unlock(&neighbourBase->lock);
    node->valid = false;
    avlatree_unlock(&node->lock);
    avlEnqueueInFreeList(set, baseNodeContainer);
    avlEnqueueInFreeList(set, neighbourBaseContainer);
    avlEnqueueInFreeList(set, parentRoute);
#ifdef AVLATREE_COUNT_ROUTE_NODES
    int count = atomic_fetch_sub(&set->nrOfRouteNodes, 1);
    printf("%d\n", count);
#endif
    return true;
}


static inline
bool aat_set_insert_opt(AVLATreeSet * set,
                       void * valuePtr,
                       unsigned int valueSize,
                       int keyPosition,
                       bool overwrite,
                       void * (*extract_key)(void * v, int keyPos),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       bool copyIn,
                       void ** replacedWriteBack,
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    void * key = AVLATREE_EXTRACT_KEY_UNTRANSFORMED(valuePtr);
    //Find base node
    AVLBaseOrAVLRouteNode * currentNode = set->root;
    AVLBaseOrAVLRouteNode * prevNode = set->root;
    AVLRouteNode * currentAVLRouteNode = NULL;
    while(!currentNode->isAVLBaseNode){
        currentAVLRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentAVLRouteNode->value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentAVLRouteNode->left;
        } else {
            currentNode = currentAVLRouteNode->right;
        }
    }
    //Lock and handle contention
    AVLBaseNode * node = &currentNode->baseOrRoute.base;
    AvlatreeLock * lock = &node->lock;
    void ** parent = NULL;
    bool contention = true;
    avlatree_lock(lock);
    if(!node->valid){
        //Retry
        avlatree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        return aat_set_insert_opt(set,
                                 valuePtr,
                                 valueSize,
                                 keyPosition,
                                 overwrite,
                                 extract_key,
                                 are_equal,
                                 less_than,
                                 copyIn,
                                 replacedWriteBack,
                                 custom_free,
                                 freeContext);
    } else if(lock->statistics > AVLATREE_MAX_CONTENTION_STATISTICS){
        if(currentAVLRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentAVLRouteNode->left == currentNode){
            parent = (void**)&currentAVLRouteNode->left;
        }else{
            parent = (void**)&currentAVLRouteNode->right;
        }
    } else if(lock->statistics < AVLATREE_MIN_CONTENTION_STATISTICS){
        if(currentAVLRouteNode == NULL){
            parent = (void**)set->root;
        }else if(currentAVLRouteNode->left == currentNode){
            parent = (void**)currentAVLRouteNode->left;
        }else{
            parent = (void**)currentAVLRouteNode->right;
        }
        contention = false;
    }
    //Handle empty before and then normal case
    bool returnValue = false;
    if(node->root == NULL) {
        AVLValueNode * newNode = AVLATREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
        node->root = newNode;
        //node->nrOfElements = node->nrOfElements + 1;
        returnValue = true;
    } else {
        returnValue = aat_update_helper_AVLValueNode(node,
                                                 valuePtr,
                                                 valueSize,
                                                 keyPosition,
                                                 overwrite,
                                                 extract_key,
                                                 are_equal,
                                                 less_than,
                                                 replacedWriteBack,
                                                 custom_free,
                                                 freeContext);
    }
    if(parent != NULL){
        if(contention){
            if(!avlContentionReduceSplit(currentNode,
                                         parent,
                                         set,
                                         keyPosition,
                                         extract_key,
                                         less_than)){
                avlatree_unlock(lock);
            }
        }else{
            if(!avlNoContentionMerge(currentNode,
                                     prevNode,
                                     set,
                                     keyPosition,
                                     extract_key,
                                     are_equal,
                                     less_than,
                                     custom_free,
                                     freeContext)){
                avlatree_unlock(lock);
            }
        }
        return returnValue;
    }
    avlatree_unlock(lock);
    return returnValue;
}

static inline
void aat_set_insert(void * setParam, void * valuePtr, unsigned int valueSize){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    aat_set_insert_opt(set,
                      valuePtr,
                      valueSize,
                      set->keyPosition,
                      true,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      true,
                      NULL,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);
}

static inline
void aat_set_insert_ops(void * setParam, void * valuePtr, unsigned int valueSize, bool copyIn, void ** replacedWriteBack){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    aat_set_insert_opt(set,
                      valuePtr,
                      valueSize,
                      set->keyPosition,
                      true,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      copyIn,
                      replacedWriteBack,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);
}



static inline
bool aat_set_insert_new(void * setParam, void * valuePtr, unsigned int valueSize){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    bool returnValue = aat_set_insert_opt(set,
                                         valuePtr, valueSize,
                                         set->keyPosition,
                                         false,
                                         set->extract_key,
                                         set->are_equal,
                                         set->less_than,
                                         true,
                                         NULL,
                                         set->free,
                                         set->freeContext);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}


#ifndef AVLATREE_SPECIAL_COPY_OUT
#    define AVLATREE_SPECIAL_COPY_OUT_PARAMS
#    define AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS
#    define AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL
#endif

static inline
void * aat_lookup_AVLValueNode(AVLValueNode * searchNode,
                           void * key,
                           int keyPosition,
                           void * (*extract_key)(void * v, int keyPos),
                           bool (*are_equal)(void * v1, void * v2),
                           bool (*less_than)(void * v1, void * v2) AVLATREE_SPECIAL_COPY_OUT_PARAMS){
    AVLValueNode * currentNode = searchNode;
    while(currentNode != NULL){
        void * nodeKey = extract_key(currentNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->left;
        } else if (are_equal(key, nodeKey)) {
#ifdef AVLATREE_SPECIAL_COPY_OUT
                AVLATREE_SPECIAL_COPY_OUT
#else
                void * toReturn = AVLATREE_MALLOC(currentNode->valueSize);
                memcpy(toReturn, currentNode->value, currentNode->valueSize);
#endif
                return toReturn;
        } else {
            currentNode = currentNode->right;
        }
    }
    return NULL;
}



static inline
void * aat_set_lookup_opt(AVLATreeSet * set, 
                          void * key,
                          int keyPosition,
                          void * (*extract_key)(void * v, int keyPos),
                          bool (*are_equal)(void * v1, void * v2),
                          bool (*less_than)(void * v1, void * v2),
                          void (*custom_free)(void * v, void * context),
                          void * freeContext AVLATREE_SPECIAL_COPY_OUT_PARAMS){
    //Find base node
    AVLBaseOrAVLRouteNode * currentNode = set->root;
    AVLBaseOrAVLRouteNode * prevNode = set->root;
    AVLRouteNode * currentAVLRouteNode = NULL;
    while(!currentNode->isAVLBaseNode){
        currentAVLRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentAVLRouteNode->value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentAVLRouteNode->left;
        } else {
            currentNode = currentAVLRouteNode->right;
        }
    }
    //Lock and handle contention
    AVLBaseNode * node = &currentNode->baseOrRoute.base;
    AvlatreeLock * lock = &node->lock;
#ifdef AVLATREE_SEQ_LOCK
    //Try fast path first
    unsigned long validationToken = seq_start_read(&lock->lock);
    void * possibleRetValue;
    if(node->valid){
        if(node->root == NULL) {
            possibleRetValue = NULL;
        } else {
            possibleRetValue = aat_lookup_AVLValueNode(node->root,
                                                       key,
                                                       keyPosition,
                                                       extract_key,
                                                       are_equal,
                                                       less_than AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
        }
        if(seq_validate_read(&lock->lock, validationToken)){
            return possibleRetValue;//Write free read!
        }
    }
#elif defined (AVLATREE_DRMCS_LOCK)
    if(!drmcs_is_locked(&lock->lock)){
        void * retValue;
        drmcs_rlock(&lock->lock);
        if(node->root == NULL) {
            retValue = NULL;
        } else {
            retValue = aat_lookup_AVLValueNode(node->root,
                                               key,
                                               keyPosition,
                                               extract_key,
                                               are_equal,
                                               less_than AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
        }
        drmcs_runlock(&lock->lock);
        return retValue;
    }
#elif defined (HLE_FAST_READ_PATH)
    if(hle_try_lock(&lock->lock)){
        void * retValue;
        if(node->root == NULL) {
            retValue = NULL;
        } else {
            retValue = aat_lookup_AVLValueNode(node->root,
                                               key,
                                               keyPosition,
                                               extract_key,
                                               are_equal,
                                               less_than AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
        }
        hle_unlock(&lock->lock);
        return retValue;
    }
#endif
    void ** parent = NULL;
    bool contention = true;
    avlatree_lock(lock);
    if(!node->valid){
        //Retry
        avlatree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        return aat_set_lookup_opt(set,
                                  key,
                                  keyPosition,
                                  extract_key,
                                  are_equal,
                                  less_than,
                                  custom_free,
                                  freeContext AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    } else if(lock->statistics > AVLATREE_MAX_CONTENTION_STATISTICS){
        if(currentAVLRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentAVLRouteNode->left == currentNode){
            parent = (void**)&currentAVLRouteNode->left;
        }else{
            parent = (void**)&currentAVLRouteNode->right;
        }
    } else if(lock->statistics < AVLATREE_MIN_CONTENTION_STATISTICS){
        if(currentAVLRouteNode == NULL){
            parent = (void**)set->root;
        }else if(currentAVLRouteNode->left == currentNode){
            parent = (void**)currentAVLRouteNode->left;
        }else{
            parent = (void**)currentAVLRouteNode->right;
        }
        contention = false;
    }
    //Handle empty before and then normal case
    void * returnValue = NULL;
    if(node->root == NULL) {
        returnValue = NULL;
    } else {
        returnValue = aat_lookup_AVLValueNode(node->root,
                                              key,
                                              keyPosition,
                                              extract_key,
                                              are_equal,
                                              less_than AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    }
    if(parent != NULL){
        if(contention){
            if(!avlContentionReduceSplit(currentNode,
                                         parent,
                                         set,
                                         keyPosition,
                                         extract_key,
                                         less_than)){
                avlatree_unlock(lock);
            }
        } else {
            if(!avlNoContentionMerge(currentNode,
                                     prevNode,
                                     set,
                                     keyPosition,
                                     extract_key,
                                     are_equal,
                                     less_than,
                                     custom_free,
                                     freeContext)){
                avlatree_unlock(lock);
            }
        }
        return returnValue;
    }
    avlatree_unlock(lock);
    return returnValue;
}

static inline
void * aat_set_lookup(void * setParam, 
                     void * key){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    void * returnValue = aat_set_lookup_opt(set,
                                           key,
                                           set->keyPosition,
                                           set->extract_key,
                                           set->are_equal,
                                            set->less_than,
                                            set->free,
                                            set->freeContext AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}

static inline
void * aat_set_lookup_ops(void * setParam, 
                         void * key AVLATREE_SPECIAL_COPY_OUT_PARAMS){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    void * returnValue = aat_set_lookup_opt(set,
                                            key,
                                            set->keyPosition,
                                            set->extract_key,
                                            set->are_equal,
                                            set->less_than,
                                            set->free,
                                            set->freeContext AVLATREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    drmcs_runlock(&set->globalLock);
    return returnValue;
}


/*
 * Deletion helpers
 */
static int avl_balance_left(AVLValueNode **this) 
{
    AVLValueNode *p, *p1, *p2;
    int b1, b2, h = 1;
    
    p = *this;
    switch (p->balance) {
    case -1:
	p->balance = 0;
	break;
    case 0:
	p->balance = 1;
	h = 0;
	break;
    case 1:
	p1 = p->right;
	b1 = p1->balance;
	if (b1 >= 0) { /* Single RR rotation */
	    p->right = p1->left;
	    p1->left = p;
	    if (b1 == 0) {
		p->balance = 1;
		p1->balance = -1;
		h = 0;
	    } else {
		p->balance = p1->balance = 0;
	    }
	    (*this) = p1;
	} else { /* Double RL rotation */
	    p2 = p1->left;
	    b2 = p2->balance;
	    p1->left = p2->right;
	    p2->right = p1;
	    p->right = p2->left;
	    p2->left = p;
	    p->balance = (b2 == 1) ? -1 : 0;
	    p1->balance = (b2 == -1) ? 1 : 0;
	    p2->balance = 0;
	    (*this) = p2;
	}
	break;
    }
    return h;
}

static int avl_balance_right(AVLValueNode **this) 
{
    AVLValueNode *p, *p1, *p2;
    int b1, b2, h = 1;
    
    p = *this;
    switch (p->balance) {
    case 1:
	p->balance = 0;
	break;
    case 0:
	p->balance = -1;
	h = 0;
	break;
    case -1:
	p1 = p->left;
	b1 = p1->balance;
	if (b1 <= 0) { /* Single LL rotation */
	    p->left = p1->right;
	    p1->right = p;
	    if (b1 == 0) {
		p->balance = -1;
		p1->balance = 1;
		h = 0;
	    } else {
		p->balance = p1->balance = 0;
	    }
	    (*this) = p1;
	} else { /* Double LR rotation */
	    p2 = p1->right;
	    b2 = p2->balance;
	    p1->right = p2->left;
	    p2->left = p1;
	    p->left = p2->right;
	    p2->right = p;
	    p->balance = (b2 == -1) ? 1 : 0;
	    p1->balance = (b2 == 1) ? -1 : 0;
	    p2->balance = 0;
	    (*this) = p2;
	}
    }
    return h;
}


static int avl_delsub(AVLValueNode **this) 
{
    AVLValueNode **tstack[AVLATREE_STACK_NEED];
    int tpos = 0;
    AVLValueNode *q = (*this);
    AVLValueNode **r = &(q->left);
    int h;

    /*
     * Walk down the tree to the right and search 
     * for a void right child, pick that child out
     * and return it to be put in the deleted 
     * object's place.
     */
    
    while ((*r)->right != NULL) {
	tstack[tpos++] = r;
	r = &((*r)->right);
    }
    *this = *r;
    *r = (*r)->left;
    (*this)->left = q->left;
    (*this)->right = q->right;
    (*this)->balance = q->balance;
    tstack[0] = &((*this)->left);
    h = 1;
    while (tpos && h) {
	r = tstack[--tpos];
	h = avl_balance_right(r);
    }
    return h;
}

static inline
AVLValueNode * aat_set_delete_AVLValueNode(AVLBaseNode * baseNode,
                                           void * key,
                                           int keyPosition,
                                           void * (*extract_key)(void * v, int keyPos),
                                           bool (*are_equal)(void * v1, void * v2),
                                           bool (*less_than)(void * v1, void * v2),
                                           void (*custom_free)(void * v, void * context),
                                           void * freeContext,
                                           bool returnDeleted){
    (void)custom_free;
    (void)freeContext;
    AVLValueNode **tstack[AVLATREE_STACK_NEED];
    int tpos = 0;
    int dstack[AVLATREE_STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    AVLValueNode **this = &baseNode->root;
    int dir;
    AVLValueNode *q = NULL;

    /*
     * Somewhat complicated, deletion in an AVL tree,
     * The two helpers balance_left and balance_right are used to
     * keep the balance. As in insert, we do the stacking ourselves.
     */

    dstack[dpos++] = AVLATREE_DIR_END;
    for (;;) {
	if (!*this) { /* Failure */
	    return NULL;
	} else if (less_than(key, extract_key((*this)->value, keyPosition))) {
	    dstack[dpos++] = AVLATREE_DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else if((are_equal(key, extract_key((*this)->value, keyPosition)))) { /* Equal key, found the one to delete*/
	    q = (*this);
	    if (q->right == NULL) {
		(*this) = q->left;
		state = 1;
	    } else if (q->left == NULL) {
		(*this) = q->right;
		state = 1;
	    } else {
		dstack[dpos++] = AVLATREE_DIR_LEFT;
		tstack[tpos++] = this;
		state = avl_delsub(this);
	    }
	    break;
	} else { /* go right */
	    dstack[dpos++] = AVLATREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	}
    }
    while (state && ( dir = dstack[--dpos] ) != AVLATREE_DIR_END) {
	this = tstack[--tpos];
	if (dir == AVLATREE_DIR_LEFT) {
	    state = avl_balance_left(this);
	} else {
	    state = avl_balance_right(this);
	}
    }
    if(q!=NULL && !returnDeleted){
        custom_free(q, freeContext);
        return NULL;
    } else {
        return q;
    }

}

static inline
void aat_set_delete_opt(AVLATreeSet * set, 
                       void * key,
                       int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    //Find base node
    AVLBaseOrAVLRouteNode * currentNode = set->root;
    AVLBaseOrAVLRouteNode * prevNode = set->root;
    AVLRouteNode * currentAVLRouteNode = NULL;
    while(!currentNode->isAVLBaseNode){
        currentAVLRouteNode = &currentNode->baseOrRoute.route;
        void * nodeKey = extract_key(currentAVLRouteNode->value, keyPosition);
        prevNode = currentNode;
        if(less_than(key, nodeKey)) {
            currentNode = currentAVLRouteNode->left;
        } else {
            currentNode = currentAVLRouteNode->right;
        }
    }
    //Lock and handle contention
    AVLBaseNode * node = &currentNode->baseOrRoute.base;
    AvlatreeLock * lock = &node->lock;
    void ** parent = NULL;
    bool contention = true;
    avlatree_lock(lock);
    if(!node->valid){
        //Retry
        avlatree_unlock(lock);
        drmcs_runlock(&set->globalLock);
        drmcs_rlock(&set->globalLock);
        aat_set_delete_opt(set, 
                          key,
                          keyPosition,
                          extract_key,
                          are_equal,
                          less_than,
                          custom_free,
                          freeContext);
        return;
    } else if(lock->statistics > AVLATREE_MAX_CONTENTION_STATISTICS){
        if(currentAVLRouteNode == NULL){
            parent = (void**)&set->root;
        }else if(currentAVLRouteNode->left == currentNode){
            parent = (void**)&currentAVLRouteNode->left;
        }else{
            parent = (void**)&currentAVLRouteNode->right;
        }
    } else if(lock->statistics < AVLATREE_MIN_CONTENTION_STATISTICS){
        if(currentAVLRouteNode == NULL){
            parent = (void**)set->root;
        }else if(currentAVLRouteNode->left == currentNode){
            parent = (void**)currentAVLRouteNode->left;
        }else{
            parent = (void**)currentAVLRouteNode->right;
        }
        contention = false;
    }
    //Handle empty before and then normal case
    if(node->root != NULL) {
        aat_set_delete_AVLValueNode(node,
                                    key,
                                    keyPosition,
                                    extract_key,
                                    are_equal,
                                    less_than,
                                    custom_free,
                                    freeContext,
                                    false);
    }
    if(parent != NULL){
        if(contention){
            if(!avlContentionReduceSplit(currentNode,
                                         parent,
                                         set,
                                         keyPosition,
                                         extract_key,
                                         less_than)){
                avlatree_unlock(lock);
            }
        }else{
            if(!avlNoContentionMerge(currentNode,
                                     prevNode,
                                     set,
                                     keyPosition,
                                     extract_key,
                                     are_equal,
                                     less_than,
                                     custom_free,
                                     freeContext)){
                avlatree_unlock(lock);
            }
        }
        return;
    }
    avlatree_unlock(lock);
}

static inline
void aat_set_delete(void * setParam, 
                   void * key,
                   unsigned int keySize){
    (void)keySize;
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    drmcs_rlock(&set->globalLock);
    aat_set_delete_opt(set,
                      key,
                      set->keyPosition,
                      set->extract_key,
                      set->are_equal,
                      set->less_than,
                      set->free,
                      set->freeContext);
    drmcs_runlock(&set->globalLock);

}


static inline
void aat_set_free_helper_helper(AVLValueNode * node,
                               void (*custom_free)(void * v, void * context),
                               void * freeContext){
    if(node!=NULL){
        aat_set_free_helper_helper(node->left, custom_free, freeContext);
        aat_set_free_helper_helper(node->right, custom_free, freeContext);
        custom_free(node, freeContext);
    }
}

static inline
void aat_set_free_helper(AVLBaseOrAVLRouteNode * root,
                        void (*custom_free)(void * v, void * freeContext),
                        void * freeContext){
    if(root->isAVLBaseNode){
        aat_set_free_helper_helper(root->baseOrRoute.base.root, custom_free, freeContext);
        AVLATREE_FREE(root);
    }else{
        AVLRouteNode * route = &root->baseOrRoute.route;
        aat_set_free_helper(route->left, custom_free, freeContext);
        aat_set_free_helper(route->right, custom_free, freeContext);
        free(route->lock);
        AVLATREE_FREE(root);
    }
}

static inline
void aat_set_destroy(void * setParam){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    for(int i = 0; i < AVLATREE_FREE_LIST_SIZE; i++){
        AVLBaseOrAVLRouteNode * elementToFree = (AVLBaseOrAVLRouteNode *)atomic_load(&set->freeList[i]);
        if(elementToFree != NULL){
            if(elementToFree->isAVLBaseNode){
                AVLATREE_LOCK_DESTROY_FUNCTION(&elementToFree->baseOrRoute.base.lock.lock); //Free deleted base
                AVLATREE_FREE(elementToFree);
            }else{
                free(elementToFree->baseOrRoute.route.lock); //Free deleted route
                AVLATREE_FREE(elementToFree);   
            }
        }
    }
    aat_set_free_helper(set->root, set->free, set->freeContext);
}

static inline
void aat_set_free(void * setParam){
    aat_set_destroy(setParam);
    AVLATREE_FREE(setParam);
}

static inline
unsigned int aat_set_size(AVLATreeSet * set){
    (void)set;
    //TODO Fix code for this!
    return 0;
}


static inline
void aat_set_print_helper_helper(AVLValueNode * node, char * (*to_string)(void * v1)){
    if(node==NULL){
        printf("NULL");
    }else{
        printf("V(");
        char * valueString = to_string(node->value);
        printf("%s", valueString);
        AVLATREE_FREE(valueString);
        printf(",");
        aat_set_print_helper_helper(node->left, to_string);
        printf(",");
        aat_set_print_helper_helper(node->right, to_string);
        printf(")");
    }
}

static inline
void aat_set_print_helper(AVLBaseOrAVLRouteNode * root, char * (*to_string)(void * v1)){
    if(root->isAVLBaseNode){
        printf("B(");
        aat_set_print_helper_helper(root->baseOrRoute.base.root, to_string);
        printf(")");
    }else{
        AVLRouteNode * route = &root->baseOrRoute.route;
        printf("R(");
        char * valueString = to_string(route->value);
        printf("%s", valueString);
        AVLATREE_FREE(valueString);
        printf(",");
        aat_set_print_helper(route->left, to_string);
        printf(",");
        aat_set_print_helper(route->right, to_string);
        printf(")");
    }
}

static inline
void aat_set_print_helper_helper_dot(AVLValueNode * node, char * (*to_string)(void * v1)){
    if(node!=NULL){
        if(node->left !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            AVLATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->left->value);
            printf("%s", valueString);
            AVLATREE_FREE(valueString);
            printf(";\n");
        }
        if(node->right !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            AVLATREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->right->value);
            printf("%s", valueString);
            AVLATREE_FREE(valueString);
            printf(";\n");
        }
        aat_set_print_helper_helper_dot(node->left, to_string);
        aat_set_print_helper_helper_dot(node->right, to_string);
    }
}

static inline
void aat_set_print_helper_dot(AVLBaseOrAVLRouteNode * root, char * (*to_string)(void * v1)){
    if(root->isAVLBaseNode){
        printf("B(\n");
        printf("digraph G{\n");
        printf("  graph [ordering=\"out\"];\n");
        aat_set_print_helper_helper_dot(root->baseOrRoute.base.root, to_string);
        printf("}\n");
        printf("\n)");
    }else{
        AVLRouteNode * route = &root->baseOrRoute.route;
        printf("R(");
        char * valueString = to_string(route->value);
        printf("%s", valueString);
        AVLATREE_FREE(valueString);
        printf(",");
        aat_set_print_helper_dot(route->left, to_string);
        printf(",");
        aat_set_print_helper_dot(route->right, to_string);
        printf(")");
    }
}


static inline
void aat_set_print(void * setParam){
    AVLATreeSet * set = (AVLATreeSet *)setParam;
    aat_set_print_helper(set->root, set->to_string);
    printf("\n\nDOT:\n\n");
    aat_set_print_helper_dot(set->root, set->to_string);
    printf("\n\n");
}

static inline
char * aat_set_to_string(void * setParam){
    (void)setParam;
    //DODO FIX CODE FOR THIS!
    return NULL;
}

static inline
bool aat_set_is_concurrent(){
    return true;
}

static inline
OOSet * oo_aat_set_create(unsigned int keyPosition,
                         void * (*extract_key)(void * v, int keyPos),
                         unsigned int (*hash_key)(void * k),
                         bool (*are_equal)(void * v1, void * v2),
                         bool (*less_than)(void * v1, void * v2),
                         char * (*to_string)(void * v1)){
    AVLATreeSet * set = plain_aat_set_create(keyPosition,
                                         extract_key,
                                         hash_key,
                                         are_equal,
                                         less_than,
                                         to_string);
    OOSet * ooset = AVLATREE_MALLOC(sizeof(OOSet));
    ooset->set = set;
    ooset->m = &AVLATREE_SET_METHOD_TABLE;
    return ooset;
}


#endif
