#ifndef LTREE_SET_H
#define LTREE_SET_H

/*
 * The LT tree related code in this file is derived from the
 * Erlang/OTP ordered_set implementation that was originally written
 * by Patrik Nyblom (https://github.com/bufflig).
 *
 * The LT tree implementation in Erlang/OTP can be found here:
 * https://github.com/erlang/otp/blob/75937381ec52e1714cfbc5936441b15bb1240bdb/erts/emulator/beam/erl_db_tree.c
 *
 * The following copyright apply to the code derived from Erlang/OTP:
 */

/*
 * %CopyrightBegin%
 *
 * Copyright Ericsson AB 1998-2013. All Rights Reserved.
 *
 * The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved online at http://www.erlang.org/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * %CopyrightEnd%
 */

#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "misc/thread_includes.h"
#include "misc/padded_types.h"
#include "data_structures/oo_set_interface.h"
#include "locks/drmcs_lock.h"
#include "locks/tatas_lock.h"
#include "locks/seq_lock.h"
#ifdef USE_HLE_LOCK
#   include "locks/experimental/hle_lock.h"
#endif

#ifndef LTREE_MALLOC
#      define LTREE_MALLOC(size) malloc(size)
#endif

#ifndef LTREE_FREE
#      define LTREE_FREE(data) free(data)
#endif

#ifndef LTREE_NEW_VALUE_NODE
#      define LTREE_NEW_VALUE_NODE(size, value, context) ltAllocateValueNode(size, value)
#endif

#ifndef LTREE_REPLACE_VALUE_NODE
#      define LTREE_REPLACE_VALUE_NODE(size, oldValue, value, custom_free, context) ltReplaceValueNode(size,value,oldValue,custom_free,freeContext)
#endif

#ifndef LTREE_EXTRACT_KEY_UNTRANSFORMED
#      define LTREE_EXTRACT_KEY_UNTRANSFORMED(untransformed) extract_key(valuePtr, keyPosition)
#endif

//#define LTREE_COUNT_ROUTE_NODES 1

//#define LTREE_PTHREAD_LOCK 1
//#define LTREE_TATAS_LOCK 1
//#define LTREE_SEQ_LOCK 1
#define LTREE_PTHREAD 1
#define LTREE_TATAS 2
#define LTREE_SEQ 3
#define LTREE_HLE 4
#define LTREE_DRMCS 5

#ifndef LTREE_GLOBAL_LOCK_TYPE
#    define LTREE_PTHREAD_LOCK 1
#elif LTREE_GLOBAL_LOCK_TYPE == LTREE_PTHREAD
#    define LTREE_PTHREAD_LOCK 1
#elif LTREE_GLOBAL_LOCK_TYPE == LTREE_TATAS
#    define LTREE_TATAS_LOCK 1
#elif LTREE_GLOBAL_LOCK_TYPE == LTREE_SEQ
#    define LTREE_SEQ_LOCK 1
#elif LTREE_GLOBAL_LOCK_TYPE == LTREE_HLE
#    define LTREE_HLE_LOCK 1
#elif LTREE_GLOBAL_LOCK_TYPE == LTREE_DRMCS
#    define LTREE_DRMCS_LOCK 1
#endif

#ifdef LTREE_PTHREAD_LOCK
#define LTREE_LOCK_TYPE pthread_mutex_t
#define LTREE_LOCK_INITIALIZER(lock_ptr)                           \
    pthread_mutex_init(lock_ptr, NULL);

    /* pthread_mutexattr_t attr;                                         \ */
    /* pthread_mutexattr_init(&attr);                                    \ */
    /* pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);       \ */
    /* pthread_mutexattr_destroy(&attr); */
#define LTREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) (!pthread_mutex_trylock(lock_ptr))
#define LTREE_LOCK_LOCK_FUNCTION(lock_ptr) pthread_mutex_lock(lock_ptr)
#define LTREE_LOCK_UNLOCK_FUNCTION(lock_ptr) pthread_mutex_unlock(lock_ptr)
#define LTREE_LOCK_RLOCK_FUNCTION(lock_ptr) pthread_mutex_lock(lock_ptr)
#define LTREE_LOCK_RUNLOCK_FUNCTION(lock_ptr) pthread_mutex_unlock(lock_ptr)
#define LTREE_LOCK_DESTROY_FUNCTION(lock_ptr) pthread_mutex_destroy(lock_ptr)
#endif


#ifdef LTREE_TATAS_LOCK
#define LTREE_LOCK_TYPE TATASLock
#define LTREE_LOCK_INITIALIZER(lock_ptr) tatas_initialize(lock_ptr);
#define LTREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) tatas_try_lock(lock_ptr)
#define LTREE_LOCK_LOCK_FUNCTION(lock_ptr) tatas_lock(lock_ptr)
#define LTREE_LOCK_UNLOCK_FUNCTION(lock_ptr) tatas_unlock(lock_ptr)
#define LTREE_LOCK_RLOCK_FUNCTION(lock_ptr) tatas_lock(lock_ptr)
#define LTREE_LOCK_RUNLOCK_FUNCTION(lock_ptr) tatas_unlock(lock_ptr)
#define LTREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif

#ifdef LTREE_DRMCS_LOCK
#define LTREE_LOCK_TYPE DRMCSLock
#define LTREE_LOCK_INITIALIZER(lock_ptr) drmcs_initialize(lock_ptr);
#define LTREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) drmcs_try_lock(lock_ptr)
#define LTREE_LOCK_LOCK_FUNCTION(lock_ptr) drmcs_lock(lock_ptr)
#define LTREE_LOCK_UNLOCK_FUNCTION(lock_ptr) drmcs_unlock(lock_ptr)
#define LTREE_LOCK_RLOCK_FUNCTION(lock_ptr) drmcs_rlock(lock_ptr)
#define LTREE_LOCK_RUNLOCK_FUNCTION(lock_ptr) drmcs_runlock(lock_ptr)
#define LTREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif

#ifdef LTREE_SEQ_LOCK
#define LTREE_LOCK_TYPE SEQLock
#define LTREE_LOCK_INITIALIZER(lock_ptr) seq_initialize(lock_ptr);
#define LTREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) seq_try_lock(lock_ptr)
#define LTREE_LOCK_LOCK_FUNCTION(lock_ptr) seq_lock(lock_ptr)
#define LTREE_LOCK_UNLOCK_FUNCTION(lock_ptr) seq_unlock(lock_ptr)
#define LTREE_LOCK_RLOCK_FUNCTION(lock_ptr) seq_lock(lock_ptr)
#define LTREE_LOCK_RUNLOCK_FUNCTION(lock_ptr) seq_unlock(lock_ptr)
#define LTREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif


#ifdef LTREE_HLE_LOCK
#define LTREE_LOCK_TYPE HLELock
#define LTREE_LOCK_INITIALIZER(lock_ptr) hle_initialize(lock_ptr)
#define LTREE_LOCK_TRY_LOCK_FUNCTION(lock_ptr) hle_try_lock(lock_ptr)
#define LTREE_LOCK_LOCK_FUNCTION(lock_ptr) hle_lock(lock_ptr)
#define LTREE_LOCK_UNLOCK_FUNCTION(lock_ptr) hle_unlock(lock_ptr)
#define LTREE_LOCK_RLOCK_FUNCTION(lock_ptr) hle_lock(lock_ptr)
#define LTREE_LOCK_RUNLOCK_FUNCTION(lock_ptr) hle_unlock(lock_ptr)
#define LTREE_LOCK_DESTROY_FUNCTION(lock_ptr) ;
#endif



#define LTREE_REBUILD_CONSTANT 2
#define LTREE_STACK_NEED 50
#define LTREE_DIR_LEFT 0
#define LTREE_DIR_RIGHT 1
#define LTREE_DIR_END 2 



typedef struct LTValueNodeImpl{
    struct LTValueNodeImpl * left;
    struct LTValueNodeImpl * right;
    int balance;
    unsigned int valueSize;
    char value[];
} LTValueNode;


typedef struct {
    LTREE_LOCK_TYPE globalLock;
    LTValueNode * root;
    unsigned int keyPosition;
    void * (*extract_key)(void * v, int keyPos);
    unsigned int (*hash_key)(void * k);
    bool (*are_equal)(void * v1, void * v2);
    bool (*less_than)(void * v1, void * v2);
    char * (*to_string)(void * v1);
    void (*free)(void * v, void * context);
    void * freeContext;
} LTreeSet;

extern
_Alignas(CACHE_LINE_SIZE)
OOSetMethodTable LTREE_SET_METHOD_TABLE;


static inline
LTValueNode * ltAllocateValueNode(unsigned int size, void * value){
    LTValueNode * newLTValueNode = LTREE_MALLOC(sizeof(LTValueNode) + size);
    newLTValueNode->left = NULL;
    newLTValueNode->right = NULL;
    newLTValueNode->valueSize = size;
    newLTValueNode->balance = 0;
    memcpy(newLTValueNode->value, value, size);
    return newLTValueNode;
}

static inline
LTValueNode * ltReplaceValueNode(unsigned int size,
                                   void * value,
                                   LTValueNode * toReplace,
                                   void (*custom_free)(void * v, void * context),
                                   void * freeContext){
    if(toReplace->valueSize < size){
        LTValueNode * toReturn = ltAllocateValueNode(size, value);
        custom_free(toReplace, freeContext);
        return toReturn;
    }else{
        toReplace->left = NULL;
        toReplace->right = NULL;
        toReplace->valueSize = size;
        toReplace->balance = 0;
        memcpy(toReplace->value, value, size);
        return toReplace;
    }

}



 static inline void lt_standard_free(void * toFree, void * freeContext){
     (void)freeContext;
     LTREE_FREE(toFree);
 }

static inline
void plain_lt_set_init(LTreeSet * set,
                       unsigned int keyPosition,
                       void * (*extract_key)(void * v, int keyPos),
                       unsigned int (*hash_key)(void * k),
                       bool (*are_equal)(void * v1, void * v2),
                       bool (*less_than)(void * v1, void * v2),
                       char * (*to_string)(void * v1),
                       void (*custom_free)(void * v, void * context),
                       void * freeContext){
    LTREE_LOCK_INITIALIZER(&set->globalLock);
    set->keyPosition = keyPosition;
    set->extract_key = extract_key;
    set->hash_key = hash_key;
    set->are_equal = are_equal;
    set->less_than = less_than;
    set->to_string = to_string;
    if(custom_free == NULL){
        set->free = lt_standard_free;
        set->freeContext = NULL;
    }else{
        set->free = custom_free;
        set->freeContext = freeContext;
    }
    set->root = NULL;
}


static inline
LTreeSet * plain_lt_set_create(unsigned int keyPosition,
                               void * (*extract_key)(void * v, int keyPos),
                               unsigned int (*hash_key)(void * k),
                               bool (*are_equal)(void * v1, void * v2),
                               bool (*less_than)(void * v1, void * v2),
                               char * (*to_string)(void * v1)){
    LTreeSet * set = LTREE_MALLOC(sizeof(LTreeSet));
    plain_lt_set_init(set,
                      keyPosition,
                      extract_key,
                      hash_key,
                      are_equal,
                      less_than,
                      to_string,
                      NULL,
                      NULL);
    return set;
}

static inline
int lt_compare_hash_codes2(unsigned int code1, unsigned int code2){
    return code1 - code2;
}

static inline
unsigned int ltCountNodes(LTValueNode * node){ 
    if(node == NULL) return 0;
    else if(node->left == NULL ) return ltCountNodes(node->right) + 1;
    else if(node->right == NULL ) return ltCountNodes(node->left) + 1;
    else return ltCountNodes(node->left) + 1 + ltCountNodes(node->right);
}


static inline
LTValueNode * ltInsertLTValueNode(LTValueNode * insertTo,
                                     LTValueNode * insertValueNode,
                                     int keyPosition,
                                     void * (*extract_key)(void * v, int keyPos),
                                     bool (*less_than)(void * v1, void * v2)){
    /* Non recursive insertion in LT tree, building our own stack */
    LTValueNode **tstack[LTREE_STACK_NEED];
    int tpos = 0;
    int dstack[LTREE_STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    LTValueNode * base = insertTo;
    LTValueNode **this = &base;
    void * key;
    int dir;
    LTValueNode *p1, *p2, *p;

    key = extract_key(insertValueNode->value, keyPosition);

    dstack[dpos++] = LTREE_DIR_END;
    for (;;)
	if (!*this) { /* Found our place */
	    state = 1;
	    *this = insertValueNode;
	    (*this)->balance = 0;
	    (*this)->left = (*this)->right = NULL;
	    break;
	} else if (less_than(key, extract_key((*this)->value, keyPosition))) {
	    /* go lefts */
	    dstack[dpos++] = LTREE_DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else { /* go right */
	    dstack[dpos++] = LTREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	}

    while (state && ( dir = dstack[--dpos] ) != LTREE_DIR_END) {
	this = tstack[--tpos];
	p = *this;
	if (dir == LTREE_DIR_LEFT) {
	    switch (p->balance) {
	    case 1:
		p->balance = 0;
		state = 0;
		break;
	    case 0:
		p->balance = -1;
		break;
	    case -1: /* The icky case */
		p1 = p->left;
		if (p1->balance == -1) { /* Single LL rotation */
		    p->left = p1->right;
		    p1->right = p;
		    p->balance = 0;
		    (*this) = p1;
		} else { /* Double RR rotation */
		    p2 = p1->right;
		    p1->right = p2->left;
		    p2->left = p1;
		    p->left = p2->right;
		    p2->right = p;
		    p->balance = (p2->balance == -1) ? +1 : 0;
		    p1->balance = (p2->balance == 1) ? -1 : 0;
		    (*this) = p2;
		}
		(*this)->balance = 0;
		state = 0;
		break;
	    }
	} else { /* dir == DIR_RIGHT */
	    switch (p->balance) {
	    case -1:
		p->balance = 0;
		state = 0;
		break;
	    case 0:
		p->balance = 1;
		break;
	    case 1:
		p1 = p->right;
		if (p1->balance == 1) { /* Single RR rotation */
		    p->right = p1->left;
		    p1->left = p;
		    p->balance = 0;
		    (*this) = p1;
		} else { /* Double RL rotation */
		    p2 = p1->left;
		    p1->left = p2->right;
		    p2->right = p1;
		    p->right = p2->left;
		    p2->left = p;
		    p->balance = (p2->balance == 1) ? -1 : 0;
		    p1->balance = (p2->balance == -1) ? 1 : 0;
		    (*this) = p2;
		}
		(*this)->balance = 0; 
		state = 0;
		break;
	    }
	}
    }
    return base;
}

static inline int ltValidate(LTValueNode * toTest){
    (void)toTest;
    if(toTest == NULL){
        return 0;
    }else{
        int hl = ltValidate(toTest->left);
        int hr = ltValidate(toTest->right);
        if(toTest->balance == 0 && hl != hr){
            printf("FAIL 1 %d %d\n", hl, hr);
            assert(0);
        }else if(toTest->balance == -1 && (hr - hl) != -1){
            printf("FAIL 2\n");
            assert(0);
        }else if(toTest->balance == 1 && (hr - hl) != 1){
            printf("FAIL 3\n");
            assert(0);
        }else if(toTest->balance > 1 || toTest->balance < -1){
            printf("FAIL 4\n");
            assert(0);
        }
        if(hl > hr){
            return hl + 1;
        }else{
            return hr + 1;
        }
    }
    return 0;
}


static inline
bool lt_update_helper_LTValueNode(LTreeSet * baseNode,
                                    void * valuePtr,
                                    unsigned int valueSize,
                                    int keyPosition,
                                    bool overwrite,
                                    void * (*extract_key)(void * v, int keyPos),
                                    bool (*are_equal)(void * v1, void * v2),
                                    bool (*less_than)(void * v1, void * v2),
                                    void ** replacedWriteBack,
                                    void (*custom_free)(void * v, void * context),
                                    void * freeContext){
    (void)replacedWriteBack;
    (void)custom_free;
    (void)freeContext;
    /* Non recursive insertion in LT tree, building our own stack */
    LTValueNode **tstack[LTREE_STACK_NEED];
    int tpos = 0;
    int dstack[LTREE_STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    LTValueNode **this = &baseNode->root;
    void * key;
    int dir;
    LTValueNode *p1, *p2, *p;

    key = LTREE_EXTRACT_KEY_UNTRANSFORMED(valuePtr);

    dstack[dpos++] = LTREE_DIR_END;
    for (;;)
	if (!*this) { /* Found our place */
	    state = 1;
            *this = LTREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
	    (*this)->balance = 0;
	    (*this)->left = (*this)->right = NULL;
	    break;
	} else if (less_than(key, extract_key((*this)->value, keyPosition))) {
	    /* go lefts */
	    dstack[dpos++] = LTREE_DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else if (are_equal(key, extract_key((*this)->value, keyPosition))) { 
            if(overwrite){
                LTValueNode * old = *this;
                LTValueNode * oldLeft = old->left;
                LTValueNode * oldRight = old->right;
                int oldBalance = old->balance;
                *this = LTREE_REPLACE_VALUE_NODE(valueSize, old, valuePtr, custom_free, freeContext);
                (*this)->left = oldLeft;
                (*this)->right = oldRight;
                (*this)->balance = oldBalance;
            }else{
                return false;
            }
	    break;
	} else { /* go right */
	    dstack[dpos++] = LTREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	}

    while (state && ( dir = dstack[--dpos] ) != LTREE_DIR_END) {
        this = tstack[--tpos];
        p = *this;
        if (dir == LTREE_DIR_LEFT) {
            switch (p->balance) {
            case 1:
        	p->balance = 0;
        	state = 0;
        	break;
            case 0:
        	p->balance = -1;
        	break;
            case -1: /* The icky case */
        	p1 = p->left;
        	if (p1->balance == -1) { /* Single LL rotation */
        	    p->left = p1->right;
        	    p1->right = p;
        	    p->balance = 0;
        	    (*this) = p1;
        	} else { /* Double RR rotation */
        	    p2 = p1->right;
        	    p1->right = p2->left;
        	    p2->left = p1;
        	    p->left = p2->right;
        	    p2->right = p;
        	    p->balance = (p2->balance == -1) ? +1 : 0;
        	    p1->balance = (p2->balance == 1) ? -1 : 0;
        	    (*this) = p2;
        	}
        	(*this)->balance = 0;
        	state = 0;
        	break;
            }
        } else { /* dir == DIR_RIGHT */
            switch (p->balance) {
            case -1:
        	p->balance = 0;
        	state = 0;
        	break;
            case 0:
        	p->balance = 1;
        	break;
            case 1:
        	p1 = p->right;
        	if (p1->balance == 1) { /* Single RR rotation */
        	    p->right = p1->left;
        	    p1->left = p;
        	    p->balance = 0;
        	    (*this) = p1;
        	} else { /* Double RL rotation */
        	    p2 = p1->left;
        	    p1->left = p2->right;
        	    p2->right = p1;
        	    p->right = p2->left;
        	    p2->left = p;
        	    p->balance = (p2->balance == 1) ? -1 : 0;
        	    p1->balance = (p2->balance == -1) ? 1 : 0;
        	    (*this) = p2;
        	}
        	(*this)->balance = 0;
        	state = 0;
        	break;
            }
        }
    }
    return true;
}


static inline
void lt_set_insert(void * setParam, void * valuePtr, unsigned int valueSize){
    LTreeSet * set = (LTreeSet *)setParam;
    LTREE_LOCK_LOCK_FUNCTION(&set->globalLock);
    if(set->root == NULL) {
        LTValueNode * newNode = LTREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
        set->root = newNode;
        //set->nrOfElements = node->nrOfElements + 1;
    } else {
        lt_update_helper_LTValueNode(set,
                                     valuePtr,
                                     valueSize,
                                     set->keyPosition,
                                     true,
                                     set->extract_key,
                                     set->are_equal,
                                     set->less_than,
                                     NULL,
                                     set->free,
                                     set->freeContext);
    }
    LTREE_LOCK_UNLOCK_FUNCTION(&set->globalLock);
}



static inline
bool lt_set_insert_new(void * setParam, void * valuePtr, unsigned int valueSize){
    LTreeSet * set = (LTreeSet *)setParam;
    LTREE_LOCK_LOCK_FUNCTION(&set->globalLock);
    bool returnValue = false;
    if(set->root == NULL) {
        LTValueNode * newNode = LTREE_NEW_VALUE_NODE(valueSize, valuePtr, freeContext);
        set->root = newNode;
        //set->nrOfElements = node->nrOfElements + 1;
        returnValue = true;
    } else {
        returnValue = lt_update_helper_LTValueNode(set,
                                                    valuePtr,
                                                    valueSize,
                                                    set->keyPosition,
                                                    false,
                                                    set->extract_key,
                                                    set->are_equal,
                                                    set->less_than,
                                                    NULL,
                                                    set->free,
                                                    set->freeContext);
    }
    LTREE_LOCK_UNLOCK_FUNCTION(&set->globalLock);
    return returnValue;
}


#ifndef LTREE_SPECIAL_COPY_OUT
#    define LTREE_SPECIAL_COPY_OUT_PARAMS
#    define LTREE_SPECIAL_COPY_OUT_PARAMS_PASS
#    define LTREE_SPECIAL_COPY_OUT_PARAMS_PASS_NULL
#endif

static inline
void * lt_lookup_LTValueNode(LTValueNode * searchNode,
                           void * key,
                           int keyPosition,
                           void * (*extract_key)(void * v, int keyPos),
                           bool (*are_equal)(void * v1, void * v2),
                           bool (*less_than)(void * v1, void * v2) LTREE_SPECIAL_COPY_OUT_PARAMS){
    LTValueNode * currentNode = searchNode;
    while(currentNode != NULL){
        void * nodeKey = extract_key(currentNode->value, keyPosition);
        if(less_than(key, nodeKey)) {
            currentNode = currentNode->left;
        } else if (are_equal(key, nodeKey)) {
#ifdef LTREE_SPECIAL_COPY_OUT
                LTREE_SPECIAL_COPY_OUT
#else
                void * toReturn = LTREE_MALLOC(currentNode->valueSize);
                memcpy(toReturn, currentNode->value, currentNode->valueSize);
#endif
                return toReturn;
        } else {
            currentNode = currentNode->right;
        }
    }
    return NULL;
}




static inline
void * lt_set_lookup(void * setParam, 
                     void * key){
    LTreeSet * set = (LTreeSet *)setParam;
    LTREE_LOCK_RLOCK_FUNCTION(&set->globalLock);
    void * returnValue;
    if(set->root == NULL) {
        returnValue = NULL;
    } else {
        returnValue = lt_lookup_LTValueNode(set->root,
                                            key,
                                            set->keyPosition,
                                            set->extract_key,
                                            set->are_equal,
                                            set->less_than LTREE_SPECIAL_COPY_OUT_PARAMS_PASS);
    }
    LTREE_LOCK_RUNLOCK_FUNCTION(&set->globalLock);
    return returnValue;
}


/*
 * Deletion helpers
 */
static int lt_balance_left(LTValueNode **this) 
{
    LTValueNode *p, *p1, *p2;
    int b1, b2, h = 1;
    
    p = *this;
    switch (p->balance) {
    case -1:
	p->balance = 0;
	break;
    case 0:
	p->balance = 1;
	h = 0;
	break;
    case 1:
	p1 = p->right;
	b1 = p1->balance;
	if (b1 >= 0) { /* Single RR rotation */
	    p->right = p1->left;
	    p1->left = p;
	    if (b1 == 0) {
		p->balance = 1;
		p1->balance = -1;
		h = 0;
	    } else {
		p->balance = p1->balance = 0;
	    }
	    (*this) = p1;
	} else { /* Double RL rotation */
	    p2 = p1->left;
	    b2 = p2->balance;
	    p1->left = p2->right;
	    p2->right = p1;
	    p->right = p2->left;
	    p2->left = p;
	    p->balance = (b2 == 1) ? -1 : 0;
	    p1->balance = (b2 == -1) ? 1 : 0;
	    p2->balance = 0;
	    (*this) = p2;
	}
	break;
    }
    return h;
}

static int lt_balance_right(LTValueNode **this) 
{
    LTValueNode *p, *p1, *p2;
    int b1, b2, h = 1;
    
    p = *this;
    switch (p->balance) {
    case 1:
	p->balance = 0;
	break;
    case 0:
	p->balance = -1;
	h = 0;
	break;
    case -1:
	p1 = p->left;
	b1 = p1->balance;
	if (b1 <= 0) { /* Single LL rotation */
	    p->left = p1->right;
	    p1->right = p;
	    if (b1 == 0) {
		p->balance = -1;
		p1->balance = 1;
		h = 0;
	    } else {
		p->balance = p1->balance = 0;
	    }
	    (*this) = p1;
	} else { /* Double LR rotation */
	    p2 = p1->right;
	    b2 = p2->balance;
	    p1->right = p2->left;
	    p2->left = p1;
	    p->left = p2->right;
	    p2->right = p;
	    p->balance = (b2 == -1) ? 1 : 0;
	    p1->balance = (b2 == 1) ? -1 : 0;
	    p2->balance = 0;
	    (*this) = p2;
	}
    }
    return h;
}


static int lt_delsub(LTValueNode **this) 
{
    LTValueNode **tstack[LTREE_STACK_NEED];
    int tpos = 0;
    LTValueNode *q = (*this);
    LTValueNode **r = &(q->left);
    int h;

    /*
     * Walk down the tree to the right and search 
     * for a void right child, pick that child out
     * and return it to be put in the deleted 
     * object's place.
     */
    
    while ((*r)->right != NULL) {
	tstack[tpos++] = r;
	r = &((*r)->right);
    }
    *this = *r;
    *r = (*r)->left;
    (*this)->left = q->left;
    (*this)->right = q->right;
    (*this)->balance = q->balance;
    tstack[0] = &((*this)->left);
    h = 1;
    while (tpos && h) {
	r = tstack[--tpos];
	h = lt_balance_right(r);
    }
    return h;
}

static inline
LTValueNode * lt_set_delete_LTValueNode(LTreeSet * baseNode,
                                           void * key,
                                           int keyPosition,
                                           void * (*extract_key)(void * v, int keyPos),
                                           bool (*are_equal)(void * v1, void * v2),
                                           bool (*less_than)(void * v1, void * v2),
                                           void (*custom_free)(void * v, void * context),
                                           void * freeContext,
                                           bool returnDeleted){
    (void)custom_free;
    (void)freeContext;
    LTValueNode **tstack[LTREE_STACK_NEED];
    int tpos = 0;
    int dstack[LTREE_STACK_NEED+1];
    int dpos = 0;
    int state = 0;
    LTValueNode **this = &baseNode->root;
    int dir;
    LTValueNode *q = NULL;

    /*
     * Somewhat complicated, deletion in an LT tree,
     * The two helpers balance_left and balance_right are used to
     * keep the balance. As in insert, we do the stacking ourselves.
     */

    dstack[dpos++] = LTREE_DIR_END;
    for (;;) {
	if (!*this) { /* Failure */
	    return NULL;
	} else if (less_than(key, extract_key((*this)->value, keyPosition))) {
	    dstack[dpos++] = LTREE_DIR_LEFT;
	    tstack[tpos++] = this;
	    this = &((*this)->left);
	} else if((are_equal(key, extract_key((*this)->value, keyPosition)))) { /* Equal key, found the one to delete*/
	    q = (*this);
	    if (q->right == NULL) {
		(*this) = q->left;
		state = 1;
	    } else if (q->left == NULL) {
		(*this) = q->right;
		state = 1;
	    } else {
		dstack[dpos++] = LTREE_DIR_LEFT;
		tstack[tpos++] = this;
		state = lt_delsub(this);
	    }
	    break;
	} else { /* go right */
	    dstack[dpos++] = LTREE_DIR_RIGHT;
	    tstack[tpos++] = this;
	    this = &((*this)->right);
	}
    }
    while (state && ( dir = dstack[--dpos] ) != LTREE_DIR_END) {
	this = tstack[--tpos];
	if (dir == LTREE_DIR_LEFT) {
	    state = lt_balance_left(this);
	} else {
	    state = lt_balance_right(this);
	}
    }
    if(q!=NULL && !returnDeleted){
        custom_free(q, freeContext);
        return NULL;
    } else {
        return q;
    }

}


static inline
void lt_set_delete(void * setParam, 
                   void * key,
                   unsigned int keySize){
    (void)keySize;
    LTreeSet * set = (LTreeSet *)setParam;
    LTREE_LOCK_LOCK_FUNCTION(&set->globalLock);
    if(set->root != NULL) {
        lt_set_delete_LTValueNode(set,
                                  key,
                                  set->keyPosition,
                                  set->extract_key,
                                  set->are_equal,
                                  set->less_than,
                                  set->free,
                                  set->freeContext,
                                  false);
    }
    LTREE_LOCK_UNLOCK_FUNCTION(&set->globalLock);

}


static inline
void lt_set_free_helper(LTValueNode * node,
                        void (*custom_free)(void * v, void * context),
                        void * freeContext){
    if(node!=NULL){
        lt_set_free_helper(node->left, custom_free, freeContext);
        lt_set_free_helper(node->right, custom_free, freeContext);
        custom_free(node, freeContext);
    }
}


static inline
void lt_set_destroy(void * setParam){
    LTreeSet * set = (LTreeSet *)setParam;
    lt_set_free_helper(set->root, set->free, set->freeContext);
}

static inline
void lt_set_free(void * setParam){
    lt_set_destroy(setParam);
    LTREE_FREE(setParam);
}

static inline
unsigned int lt_set_size(LTreeSet * set){
    (void)set;
    //TODO Fix code for this!
    return 0;
}


static inline
void lt_set_print_helper(LTValueNode * node, char * (*to_string)(void * v1)){
    if(node==NULL){
        printf("NULL");
    }else{
        printf("V(");
        char * valueString = to_string(node->value);
        printf("%s", valueString);
        LTREE_FREE(valueString);
        printf(",");
        lt_set_print_helper(node->left, to_string);
        printf(",");
        lt_set_print_helper(node->right, to_string);
        printf(")");
    }
}


static inline
void lt_set_print_helper_dot(LTValueNode * node, char * (*to_string)(void * v1)){
    if(node!=NULL){
        if(node->left !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            LTREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->left->value);
            printf("%s", valueString);
            LTREE_FREE(valueString);
            printf(";\n");
        }
        if(node->right !=NULL){
            char * valueString = to_string(node->value);
            printf("%s", valueString);
            LTREE_FREE(valueString);
            printf(" -> ");
            valueString = to_string(node->right->value);
            printf("%s", valueString);
            LTREE_FREE(valueString);
            printf(";\n");
        }
        lt_set_print_helper_dot(node->left, to_string);
        lt_set_print_helper_dot(node->right, to_string);
    }
}


static inline
void lt_set_print(void * setParam){
    LTreeSet * set = (LTreeSet *)setParam;
    lt_set_print_helper(set->root, set->to_string);
    printf("\n\nDOT:\n\n");
    lt_set_print_helper_dot(set->root, set->to_string);
    printf("\n\n");
}

static inline
char * lt_set_to_string(void * setParam){
    (void)setParam;
    //DODO FIX CODE FOR THIS!
    return NULL;
}

static inline
bool lt_set_is_concurrent(){
    return true;
}

static inline
OOSet * oo_lt_set_create(unsigned int keyPosition,
                         void * (*extract_key)(void * v, int keyPos),
                         unsigned int (*hash_key)(void * k),
                         bool (*are_equal)(void * v1, void * v2),
                         bool (*less_than)(void * v1, void * v2),
                         char * (*to_string)(void * v1)){
    LTreeSet * set = plain_lt_set_create(keyPosition,
                                         extract_key,
                                         hash_key,
                                         are_equal,
                                         less_than,
                                         to_string);
    OOSet * ooset = LTREE_MALLOC(sizeof(OOSet));
    ooset->set = set;
    ooset->m = &LTREE_SET_METHOD_TABLE;
    return ooset;
}


#endif
