#include "tests/test_framework.h"

#include "data_structures/sets.h"
#include "misc/random.h"

typedef union {
    unsigned int value;
    char pad[CACHE_LINE_SIZE];
} LLPaddedSeed;

typedef union {
    unsigned long value;
    char pad[CACHE_LINE_SIZE];
} LLPaddedLocalCounter;

typedef struct {
    unsigned long * localInCSCounter;
    unsigned int * localSeed;
} ThreadLocalData;


void * my_extract_key(void * v, int keyPos){
    (void)keyPos;
    return v;
}

unsigned int my_hash_key(void * keyPtr){
    int key = *((int *)keyPtr);
    return ((unsigned int)key);
}

bool my_are_equal(void * k1, void * k2){
    int  k1i = *((int *)k1);
    int k2i = *((int *)k2);
    return k1i == k2i;
}

bool my_less_than(void * k1, void * k2){
    int  k1i = *((int *)k1);
    int k2i = *((int *)k2);
    return k1i < k2i;
}

char * my_to_string(void * value){
    int  valuei = *((int *)value);
    char * str = malloc(100);
    sprintf(str, "\"%d\"", valuei);
    return str;
}

SET_TYPE * create_set(){
    return (SET_TYPE *)SET_create(SET_TYPE_NAME,
                                  0,
                                  my_extract_key,
                                  my_hash_key,
                                  my_are_equal,
                                  my_less_than,
                                  my_to_string);
}


int test_create_and_delete(){
    SET_TYPE * set = create_set();
    SET_free(set);
    return 1;
}

int test_insert(){
    SET_TYPE * set = create_set();
    int x = 10;
    SET_insert(set, &x, sizeof(int));
    void * element = SET_lookup(set, &x);
    assert(*((int *)element) == 10);
    if(SET_is_concurrent(set)){
        free(element);
    }
    SET_free(set);
    return 1;
}

int test_insert_new(){
    SET_TYPE * set = create_set();
    int x = 10;
    assert(SET_insert_new(set, &x, sizeof(int)));
    void * element = SET_lookup(set, &x);
    assert(*((int*)element) == 10);
    if(SET_is_concurrent(set)){
        free(element);
    }
    assert(!SET_insert_new(set, &x, sizeof(int)));
    SET_free(set);
    return 1;
}

int test_insert_write_over(){
    SET_TYPE * set = create_set();
    int x = 10;
    SET_insert(set, &x, sizeof(int));
    x = 5;
    SET_insert(set, &x, sizeof(int));
    x = 15;
    SET_insert(set, &x, sizeof(int));
    x = 10;
    SET_insert(set, &x, sizeof(int));
    void * element = SET_lookup(set, &x);
    assert(*((int*)element) == 10);
    if(SET_is_concurrent(set)){
        free(element);
    }
    SET_free(set);
    return 1;
}

int test_lookup(){
    SET_TYPE * set = create_set();
    int x = 42;
    SET_insert(set, &x, sizeof(int));
    void * element = SET_lookup(set, &x);
    assert(*((int*)element) == 42);
    if(SET_is_concurrent(set)){
        free(element);
    }
    SET_free(set);
    return 1;
}

int test_lookup_not_exsisting(){
    SET_TYPE * set = create_set();
    int x = 42;
    SET_insert(set, &x, sizeof(int));
    x = 43;
    assert(SET_lookup(set, &x) == NULL);
    SET_free(set);
    return 1;
}

int test_remove(){
    SET_TYPE * set = create_set();
    int x = 42;
    SET_insert(set, &x, sizeof(int));
    void * element = SET_lookup(set, &x);
    assert(*((int*)element) == 42);
    if(SET_is_concurrent(set)){
        free(element);
    }
    SET_delete(set, &x, sizeof(int));
    assert(SET_lookup(set, &x) == NULL);
    SET_free(set);
    return 1;
}

int test_remove_many(){
    int nrOfOpsOfEachType = 500;
    SET_TYPE * set = create_set();
    int elements[nrOfOpsOfEachType];
    //Create elements
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        elements[i] = i;
    }
    //Inserts
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        SET_insert(set, &elements[i], sizeof(int));
    }
    //Lookups
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        void * element = SET_lookup(set, &elements[i]);
        assert(*((int*)element) == elements[i]);
        if(SET_is_concurrent(set)){
            free(element);
        }
    }
    //Deletes
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        SET_delete(set, &elements[i], sizeof(int));
    }
    SET_free(set);
    return 1;
}

int test_insert_many(){
    int nrOfOpsOfEachType = 300;
    SET_TYPE * set = create_set();
    int elements[nrOfOpsOfEachType];
    //Create elements
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        elements[i] = i;
    }
    //Inserts
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        SET_insert(set, &elements[i], sizeof(int));
    }
    //Inserts again
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        SET_insert(set, &elements[i], sizeof(int));
    }
    //Lookups
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        void * element = SET_lookup(set, &elements[i]);
        assert(*((int*)element) == elements[i]);
        if(SET_is_concurrent(set)){
            free(element);
        }
    }
    SET_free(set);
    return 1;
}


int test_insert_lookup_delete_lookup_many(){
    int nrOfOpsOfEachType = 1024;
    SET_TYPE * set = create_set();
    int elements[nrOfOpsOfEachType];
    unsigned int seed = 13;
    //Create elements
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        elements[i] = (int)(10000*random_double(&seed));
    }
    //Inserts
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        SET_insert(set, &elements[i], sizeof(int));
    }
    //Lookups
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        void * element = SET_lookup(set, &elements[i]);
        assert(*((int*)element) == elements[i]);
        if(SET_is_concurrent(set)){
            free(element);
        }
    }
    //Inserts again!
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        assert( ! SET_insert_new(set, &elements[i], sizeof(int)));
    }
    //Lookups
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        void * element = SET_lookup(set, &elements[i]);
        assert(*((int*)element) == elements[i]);
        if(SET_is_concurrent(set)){
            free(element);
        }
    }
    //Deletes
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        SET_delete(set, &elements[i], sizeof(int));
    }
    //Lookup_again
    for(int i = 0; i < nrOfOpsOfEachType; i++){
        assert(SET_lookup(set, &elements[i]) == NULL);
    }
    SET_free(set);
    return 1;
}



LLPaddedBool stop = {.value = ATOMIC_FLAG_INIT};
LLPaddedDouble getPercentage;
LLPaddedDouble insertPercentage;
LLPaddedDouble deletePercentage;
SET_TYPE * set;

void  * critical_section_thread(void * threadLocalDataVPtr){
    ThreadLocalData * threadLocalDataPtr = (ThreadLocalData*)threadLocalDataVPtr;
    unsigned long * localInCSCounter = threadLocalDataPtr->localInCSCounter;
    unsigned int * localSeed = threadLocalDataPtr->localSeed;
    double getPercentageV = getPercentage.value;
    double getPlusInsertPercentage = getPercentageV + insertPercentage.value;
    while(!atomic_load_explicit(&stop.value, memory_order_acquire)){
        double randomNumber = random_double(localSeed);
        int element = (int)(100000*random_double(localSeed));
        if(randomNumber > getPlusInsertPercentage){
            //delete
            SET_delete(set, &element, sizeof(int));
        }if(randomNumber > getPercentageV){
            //insert
            SET_insert(set, &element, sizeof(int));
        }else {
            //get
            void * elem = SET_lookup(set, &element);
            if(SET_is_concurrent(set) && elem !=NULL){
                free(elem);
            }
        } 
        *localInCSCounter = *localInCSCounter + 1;
    }
    return NULL;
}


int test_parallel_ops(double getPercentageParm,
                      double insertPercentageParm,
                      double deletePercentageParm){
    getPercentage.value = getPercentageParm;
    insertPercentage.value = insertPercentageParm;
    deletePercentage.value = deletePercentageParm;
    set = create_set();
    struct timespec testTime= {.tv_sec = 1, .tv_nsec = 100000000}; 
    int threadCountsToTest[] = {1,2,4,8,16,32,64,128,256};
    int nrOfThreadCountsToTest = 9;
    for(int n = 0; n < nrOfThreadCountsToTest; n++){
        int i = threadCountsToTest[n];
        atomic_store(&stop.value, false);
        pthread_t threads[i];
        LLPaddedLocalCounter localInCSCounters[i];
        LLPaddedSeed localSeeds[i];
        ThreadLocalData * threadLocalData = aligned_alloc(CACHE_LINE_SIZE, sizeof(ThreadLocalData) * i);
        for(int n = 0; n < i; n++){
            localInCSCounters[n].value = 0;
            localSeeds[n].value = n;
            threadLocalData[n].localInCSCounter = &localInCSCounters[n].value;
            threadLocalData[n].localSeed = &localSeeds[n].value;
            pthread_create(&threads[n], NULL,
                           critical_section_thread,
                           &threadLocalData[n]);
        }
        nanosleep(&testTime, NULL);
        atomic_store(&stop.value, true);
        unsigned long localInCSCountersSum = 0;
        for(int n = 0; n < i; n++){
            void * resp = NULL;
            pthread_join(threads[n], &resp);
        }
        for(int n = 0; n < i; n++){
            localInCSCountersSum = localInCSCountersSum + 
                localInCSCounters[n].value;
        }
        printf("TOTAL OP COUNT %lu\n", localInCSCountersSum);
        free(threadLocalData);
        // assert(localInCSCountersSum == atomic_load(&counter.value));
    }
    SET_free(set);
    return 1;
}


int main(int argc, char **argv){
    (void)argc;
    (void)argv;
    /**/
    T(test_create_and_delete(), "test_create_and_delete()");
    T(test_insert(), "test_insert()");
    T(test_insert_write_over(), "test_insert_write_over()");
    T(test_lookup(), "test_lookup()");
    T(test_lookup_not_exsisting(), "test_lookup_not_exsisting()");
    T(test_insert_many(), "test_insert_many()");
    T(test_remove(), "test_remove()");
    T(test_remove_many(), "test_remove_many()");
    T(test_insert_lookup_delete_lookup_many(), "test_insert_lookup_delete_lookup_many()");
    /**/
    T(test_parallel_ops(0.8, 0.2, 0.0), "test_parallel_ops(0.8, 0.2, 0.0)");
    T(test_parallel_ops(0.0, 1.0, 0.0), "test_parallel_ops(0.0, 1.0, 0.0)");
    /**/
    T(test_parallel_ops(0.0, 0.2, 0.8), "test_parallel_ops(0.0, 0.2, 0.8)");
    T(test_parallel_ops(0.0, 0.7, 0.3), "test_parallel_ops(0.0, 0.7, 0.3)");
    T(test_parallel_ops(0.5, 0.4, 0.1), "test_parallel_ops(0.5, 0.4, 0.1)");
    T(test_parallel_ops(0.3, 0.4, 0.3), "test_parallel_ops(0.3, 0.4, 0.3)");
    return 0;
}

